"""
Quantiphyse - Classes for interacting with external command line programs

Copyright (c) 2013-2018 University of Oxford
"""

import os
import sys
import shlex
import subprocess
import tempfile
import re
import shutil
import logging
import six

from quantiphyse.data import load, save
from quantiphyse.utils import QpException
from quantiphyse.processes import Process

LOG = logging.getLogger(__name__)

class OutputStreamMonitor(object):
    """
    Simple file-like object which listens to the output
    of a command and sends each line to a Queue
    """
    def __init__(self, queue, suppress_duplicates=True):
        self._queue = queue
        self._suppress_duplicates = suppress_duplicates
        self._last_line = None

    def write(self, text):
        """ Handle output from the process - send each line to the queue """
        lines = text.splitlines(True)
        for line in lines:
            if line != self._last_line:
                self._queue.put(line)
            if self._suppress_duplicates:
                self._last_line = line

    def flush(self):
        """ Ignore flush requests """
        pass

def _get_files(workdir):
    """
    Get a list of files currently in a working directory
    """
    dir_files = []
    for _, _, files in os.walk(workdir):
        for dir_file in files:
            if os.path.isfile(dir_file):
                dir_files.append(dir_file)
    return dir_files

def _run_cmd(worker_id, queue, workdir, cmdline, expected_data):
    """
    Multiprocessing worker to run a command in the background
    """
    try:
        pre_run_files = _get_files(workdir)

        cmd_args = shlex.split(cmdline, posix=not sys.platform.startswith("win"))
        os.chdir(workdir)
        log = ""
        proc = subprocess.Popen(cmd_args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        while 1:
            # This returns None while subprocess is running
            retcode = proc.poll() 
            line = proc.stdout.readline()
            log += line
            if queue: queue.put(line)
            if retcode is not None: break

        if retcode != 0:
            LOG.debug("External program failed: %s", cmdline)
            LOG.debug(log)
            raise QpException("Failed to execute %s: return code %i" % (cmd_args[0], retcode), detail=log)

        post_run_files = _get_files(workdir)
        new_files = [f for f in post_run_files if f not in pre_run_files]
        LOG.debug("New files: %s", new_files)
        data = []

        for new_file in new_files:
            basename = new_file.split(".", 1)[0]
            LOG.debug("Checking if we need to output: %s %s", basename, expected_data)
            if basename in expected_data or not expected_data:
                LOG.debug("Adding output file %s", new_file)
                if basename in expected_data:
                    data.append(new_file)
                else:
                    data.append(new_file)

        return worker_id, True, (log, data)
    except Exception as exc:
        import traceback
        traceback.print_exc(exc)
        return worker_id, False, exc

def apply_backspaces(s):
    while True:
        # if you find a character followed by a backspace, remove both
        t = re.sub('.\b', '', s, count=1)
        if len(s) == len(t):
            # now remove any backspaces from beginning of string
            return re.sub('\b+', '', t)
        s = t

class LogProcess(Process):
    """
    Process which produces a log stream which can be used to 
    monitor progress
    """
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)
        self.expected_steps = []
        self.current_step = 0
        self.allow_skipping_steps = kwargs.pop("allow_skipping_steps", True)

    def timeout(self, queue):
        """ 
        Monitor queue for updates and send sig_progress 
        """
        if queue.empty(): return
        while not queue.empty():
            line = queue.get()
            # FIXME Some commands may use backspaces to show updating progress!
            self.log(line)
            self.debug(line.rstrip("\n"))
            check_step = self.current_step
            while check_step < len(self.expected_steps):
                # Expected steps can be string regexes, or tuple of (regex, description)
                expected, desc = self.expected_steps[check_step], None
                if not isinstance(expected, six.string_types):
                    expected, desc = expected

                if expected is not None and re.match(expected, line):
                    complete = float(check_step) / len(self.expected_steps)
                    self.current_step = check_step + 1
                    self.sig_progress.emit(complete)
                    if desc:
                        self.sig_step.emit(desc)
                    break
                if self.allow_skipping_steps:
                    check_step += 1

class CommandProcess(LogProcess):
    """
    Process which runs an external command in the background
    """
    def __init__(self, ivm, workdir=None, path=(), **kwargs):
        LogProcess.__init__(self, ivm, worker_fn=_run_cmd, **kwargs)
        self.path = path
        self._current_data = None
        self._current_roi = None

        if workdir is None:
            self.workdir = tempfile.mkdtemp(prefix="qp")
            self.workdir_istemp = True
        else:
            self.workdir = workdir
            self.workdir_istemp = False

    def __del__(self):
        if hasattr(self, "workdir_istemp") and self.workdir_istemp:
            try:
                shutil.rmtree(self.workdir)
            except:
                LOG.warn("Failed to remove temporary directory: %s", self.workdir)

    def add(self, data_name):
        """
        Add a data item to the working directory from the IVM
        """
        fname = os.path.join(self.workdir, data_name)
        save(self.ivm.data[data_name], fname)

    def finished(self, worker_output):
        """ 
        Add data to IVM and set log 

        Note that we need to call ``qpdata.raw()`` to make sure
        data is all loaded into memory as the files may be temporary
        """
        if self.status == Process.SUCCEEDED:
            log, data_files = worker_output[0]
            self.log(log)
            self.debug("Loading data: %s", data_files)
            for data_file in data_files:
                qpdata = load(os.path.join(self.workdir, data_file))
                qpdata.name = self.ivm.suggest_name(data_file.split(".", 1)[0], ensure_unique=False)
                qpdata.raw()
                self.ivm.add(qpdata, make_current=(data_file == self._current_data))

    def _get_cmdline(self, options):
        cmd = options.pop("cmd", None)
        if cmd is None:
            raise QpException("No command provided")

        cmdline = options.pop("cmdline", None)
        argdict = options.pop("argdict", {})
        if cmdline is None and not argdict:
            raise QpException("No command arguments provided")
            
        for arg, value in argdict.items():
            if value != "": 
                cmdline += " --%s=%s" % (arg, value)
            else:
                cmdline += " --%s" % arg
                
        cmd = self._find(cmd)
        cmdline = cmd + " " + cmdline
        LOG.debug(cmdline)
        return cmdline

    def run(self, options):
        """ 
        Run a program
        """
        cmdline = self._get_cmdline(options)

        expected_data = options.pop("output-data", [])

        self.expected_steps = options.pop("expected-steps", [None,])
        self.current_step = 0
        self._current_data = options.pop("set-current-data", None)
        self._current_roi = options.pop("set-current-roi", None)

        LOG.debug("Working directory: %s", self.workdir)
        self._add_data_from_cmdline(cmdline)

        worker_args = [self.workdir, cmdline, expected_data]
        self.start_bg(worker_args)
  
    def _find(self, cmd):
        """ 
        Find the program, either in the 'local' directory, or in $FSLDEVDIR/bin or $FSLDIR/bin 
        This is called each time the program is run so the caller can control where programs
        are searched for at any time
        """
        for bindir in self.path:
            ex = os.path.join(bindir, cmd)
            LOG.debug("Checking %s", ex)
            if os.path.isfile(ex) and os.access(ex, os.X_OK):
                return ex
            elif sys.platform.startswith("win"):
                ex += ".exe"
                if os.path.isfile(ex) and os.access(ex, os.X_OK):
                    return ex
            
        LOG.warn("Failed to find command line program: %s", cmd)
        return cmd
    
    def _add_data_from_cmdline(self, cmdline):
        """
        Add any data/roi item which match an argument to the working directory
        """
        for arg in re.split(r"\s+|=|,|\n|\t", cmdline):
            if arg in self.ivm.data:
                LOG.debug("Adding data from command line args: %s", arg)
                self.add(arg)



from .process import SmoothingProcess
from .widget import SmoothingWidget
from .tests import SmoothingWidgetTests

QP_MANIFEST = {
    "widgets" : [SmoothingWidget,],
    "widget-tests" : [SmoothingWidgetTests,],
    "processes" : [SmoothingProcess,]
}



"""
Quantiphyse - Extension to the PyQtGraph histogram to support multiple images

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import division, unicode_literals, absolute_import, print_function

from matplotlib import cm
import numpy as np

import pyqtgraph as pg

from quantiphyse.data.qpdata import remove_nans

class MultiImageHistogramWidget(pg.HistogramLUTWidget):
    """
    A histogram widget which has one array of 'source' data
    (which it gets the histogram itself and the initial levels from)
    and multiple image item views which are affected by changes to the
    levels or LUT
    """
    def __init__(self, ivl, view, *args, **kwargs):
        self.percentile = kwargs.pop("percentile", 100)
        kwargs["fillHistogram"] = False
        super(MultiImageHistogramWidget, self).__init__(*args, **kwargs)

        self.setBackground(None)
        self.ivl = ivl
        self.view = view
        self.vol = 0
        self.imgs = []
        self.updating = False

        self.ivl.sig_focus_changed.connect(self._focus_changed)
        self.view.sig_view_changed.connect(self._view_changed)
        self.sigLevelChangeFinished.connect(self._levels_changed)
        self.sigLevelsChanged.connect(self._levels_changed)
        self.sigLookupTableChanged.connect(self._lut_changed)
        self._view_changed()
        self._update_histogram()

    def add_img(self, img):
        """
        Add a pyqtgraph.ImageItem to the histogram
        """
        self.imgs.append(img)

    def remove_img(self, img):
        """
        Remove a pyqtgraph.ImageItem from the histogram
        """
        self.imgs.remove(img)

    def _view_changed(self):
        try:
            self.updating = True
            if self.view.get("cmap") != "custom":
                try:
                    self.gradient.loadPreset(self.view.get("cmap"))
                except KeyError:
                    self._set_matplotlib_gradient(self.view.get(("cmap")))
            
            if self.view.get("cmap_range") is not None:
                self.region.setRegion(self.view.get(("cmap_range")))

            #self.lut = None
            for img in self.imgs:
                img.setLookupTable(self._get_image_lut, update=True)
        finally:
            self.updating = False

    def _levels_changed(self):
        if not self.updating:
            self.view.set("cmap_range", list(self.region.getRegion()))

    def _lut_changed(self):
        if not self.updating:
            self.view.set("cmap", "custom")

    def _focus_changed(self, pos):
        if self.vol != pos[3]:
            self.vol = pos[3]
            self._update_histogram()
    
    def _update_histogram(self):
        if self.view.data is not None:
            arr = remove_nans(self.view.data.volume(self.vol))
            flat = arr.reshape([1, -1]).astype(np.float32)
            img = pg.ImageItem(flat)
            hist = img.getHistogram()
            if hist[0] is None: return
            self.plot.setData(*hist)

    def _get_image_lut(self, img):
        lut = self.getLookupTable(img, alpha=True)
        if self.view is not None:
            for row in lut:
                row[3] = self.view.get("alpha")

        #self.lut = lut
        return lut

    def _set_matplotlib_gradient(self, name):
        """
        Slightly hacky method to copy MatPlotLib gradients to pyqtgraph.

        Is not perfect because Matplotlib specifies gradients in a different way to pyqtgraph
        (specifically there is a separate list of ticks for R, G and B). So we just sample
        the colormap at 10 points which is OK for most slowly varying gradients.
        """
        cmap = getattr(cm, name)
        ticks = [(pos, [255 * v for v in cmap(pos)]) for pos in np.linspace(0, 1, 10)]
        self.gradient.restoreState({'ticks': ticks, 'mode': 'rgb'})



"""
Quantiphyse - Miscellaneous custom Qt widgets

Copyright (c) 2013-2018 University of Oxford
"""

import os
import inspect
import traceback

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets
  
from quantiphyse.processes import Process
from quantiphyse.utils import get_icon, load_matrix, local_file_from_drop_url, QpException, show_help, sf, LogSource
from quantiphyse.utils.batch import Script, to_yaml
from quantiphyse.gui.options import OptionBox, FileOption 
from quantiphyse.gui.dialogs import error_dialog, TextViewerDialog, MultiTextViewerDialog, MatrixViewerDialog
import quantiphyse.gui.dialogs

class QpWidget(QtGui.QWidget, LogSource):
    """
    Base class for a Quantiphyse widget

    The following properties are set automatically from keyword args or defaults:
      self.ivm - Image Volume Management instance
      self.ivl - ImageView instance
      self.icon - QIcon for the menu/tab
      self.name - Name for the menu
      self.description - Longer description (for tooltip)
      self.tabname - Name for the tab
    """

    def __init__(self, **kwargs):
        LogSource.__init__(self)
        QtGui.QWidget.__init__(self)
        
        # Basic metadata
        self.name = kwargs.get("name", "")
        self.tabname = kwargs.get("tabname", self.name.replace(" ", "\n"))
        self.group = kwargs.get("group", "")
        self.position = kwargs.get("position", 999)
        self.description = kwargs.get("desc", self.name)
        self.version = kwargs.get("version", "")
        self.visible = False
        self.inited = False

        # This attempts to return the directory where the derived widget is defined - 
        # so we can look there for icons as well as in the default location
        self.pkgdir = os.path.abspath(os.path.dirname(inspect.getmodule(self).__file__))
        self.icon = QtGui.QIcon(get_icon(kwargs.get("icon", ""), self.pkgdir))

        # References to core classes
        self.ivm = kwargs.get("ivm", None)
        self.ivl = kwargs.get("ivl", None)
        self.opts = kwargs.get("opts", None)
        if self.opts:
            self.opts.sig_options_changed.connect(self.options_changed)

    def get_local_file(self, name):
        """
        Get a file which is stored locally to the implementing class
        """
        return os.path.abspath(os.path.join(self.pkgdir, name))

    def init_ui(self):
        """
        Called when widget is first shown. Widgets should ideally override this to build their
        UI widgets when required, rather than in the constructor which is called at startup
        """
        pass

    def activate(self):
        """
        Called when widget is made active, so can for example connect signals to the 
        volume management or view classes, and update it's current state
        """
        pass

    def deactivate(self):
        """
        Called when widget is made inactive, so should for example disconnect signals and remove 
        any related selections from the view
        """
        pass

    def options_changed(self):
        """
        Override to respond to global option changes
        """
        pass

    def processes(self):
        """
        Get the processes to be run for this widget in its current state

        This should be implemented for widgets which run a Process or multiple 
        Processes (i.e. most widgets). Otherwise the default method should 
        be left (which throws a NotImplementedError).

        :return: Structure defining one or more processes to be run, in 
        the format expected by the batch system. This is one of the following:

          - A dictionary of Process name : Options dictionary
          - A sequence of the above
        """
        raise NotImplementedError("This widget does not support the batch process system")

class FingerTabBarWidget(QtGui.QTabBar):
    """
    Vertical tab bar used for the analysis widget setSelectionMode
    """
    def __init__(self, tab_widget, parent=None, *args, **kwargs):
        self.tab_size = QtCore.QSize(kwargs.pop('width', 100), kwargs.pop('height', 25))
        QtGui.QTabBar.__init__(self, parent, *args, **kwargs)
        self.close_icon = QtGui.QIcon(get_icon("close"))
        self.tab_widget = tab_widget

    def paintEvent(self, _):
        painter = QtGui.QStylePainter(self)
        option = QtGui.QStyleOptionTab()
 
        for index in range(self.count()):
            self.initStyleOption(option, index)
            tab_rect = self.tabRect(index)
            tab_rect.moveLeft(10)
            painter.drawControl(QtGui.QStyle.CE_TabBarTabShape, option)
            painter.drawText(tab_rect, QtCore.Qt.AlignVCenter |
                             QtCore.Qt.AlignHCenter,
                             self.tabText(index))
            painter.drawItemPixmap(tab_rect, QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter,
                                   self.tabIcon(index).pixmap(20, 20))
            w = self.tab_widget.widget(index)
            if w.group != "DEFAULT":
                tab_rect.moveLeft(-5)
                tab_rect.moveTop(tab_rect.top()+5)
                painter.drawItemPixmap(tab_rect, QtCore.Qt.AlignRight | QtCore.Qt.AlignTop,
                                       self.close_icon.pixmap(10, 10))
        painter.end()

    def mousePressEvent(self, evt):
        QtGui.QTabBar.mousePressEvent(self, evt)
        idx = self.tabAt(evt.pos())
        if idx >= 0 and evt.button() == QtCore.Qt.LeftButton:
            tab_rect = self.tabRect(idx)
            oy = evt.pos().y() - tab_rect.top() - 5
            ox = tab_rect.right() - evt.pos().x() - 5
            if ox > 0 and ox < 10 and oy > 0 and oy < 10:
                # Click was inside close button
                w = self.tab_widget.widget(idx)
                if w.group != "DEFAULT":
                    w.visible = False
                    self.tab_widget.removeTab(idx)
        
    def tabSizeHint(self, _):
        return self.tab_size

class FingerTabWidget(QtGui.QTabWidget):
    """
    A QTabWidget equivalent which uses our FingerTabBarWidget
    """
    def __init__(self, parent, *args):
        QtGui.QTabWidget.__init__(self, parent, *args)
        self.setTabBar(FingerTabBarWidget(self, width=110, height=50))
        self.setTabPosition(QtGui.QTabWidget.West)
        self.setMovable(False)
        self.setIconSize(QtCore.QSize(16, 16))
    
class HelpButton(QtGui.QPushButton):
    """
    A button for online help
    """
    def __init__(self, parent, section=""):
        super(HelpButton, self).__init__(parent)
        self.section = section
        self.setToolTip("Online Help")

        icon = QtGui.QIcon(get_icon("question-mark"))
        self.setIcon(icon)
        self.setIconSize(QtCore.QSize(14, 14))
        self.clicked.connect(self._help_clicked)

    def _help_clicked(self):
        show_help(self.section)

class BatchButton(QtGui.QPushButton):
    """
    A button which displays the batch file code for the current analysis widget
    """
    def __init__(self, widget):
        super(BatchButton, self).__init__(widget)
        self.widget = widget
        
        icon = QtGui.QIcon(get_icon("batch"))
        self.setIcon(icon)
        self.setIconSize(QtCore.QSize(14, 14))

        self.setToolTip("Show batch mode options for this widget")

        self.clicked.connect(self.show_batch_options)
        
    def show_batch_options(self):
        """
        Show a dialog box containing the batch options supplied by the parent
        """
        try:
            processes = self.widget.processes()
            text = to_yaml(processes)
            TextViewerDialog(self.widget, title="Batch options for %s" % self.widget.name, text=text).show()
        except NotImplementedError:
            # Fallback to older method
            if hasattr(self.widget, "batch_options"):
                batchopts = self.widget.batch_options()
                if len(batchopts) == 2:
                    proc_name, opts = batchopts

                    text = "  - %s:\n" % proc_name
                    text += "\n".join(["      %s: %s" % (str(k), str(v)) for k, v in opts.items()])
                    text += "\n"
                    TextViewerDialog(self.widget, title="Batch options for %s" % self.widget.name, text=text).show()
                elif len(batchopts) == 3:
                    proc_name, opts, support_files = batchopts
                    text = "  - %s:\n" % proc_name
                    text += "\n".join(["      %s: %s" % (str(k), str(v)) for k, v in opts.items()])
                    text += "\n"
                    support_files.insert(0, ("Batch code", text))
                    MultiTextViewerDialog(self.widget, title="Batch options for %s" % self.widget.name, 
                                          pages=support_files).show()     
            else:
                error_dialog("This widget does not provide a list of batch options")

class OverlayCombo(QtGui.QComboBox):
    """
    A combo box which gives a choice of data

    Really ugly hacks to make it work for data / rois
    """
    def __init__(self, ivm, parent=None, static_only=False, none_option=False, all_option=False, data=True, rois=False, **kwargs):
        super(OverlayCombo, self).__init__(parent)
        self.ivm = ivm
        self.static_only = static_only
        self.none_option = none_option
        self.all_option = all_option
        self.rois = rois
        self.data = data
        self.ivm.sig_all_data.connect(self._data_changed)

        # Whether the combo should automatically adopt the name of the
        # first data set to be added
        self._set_first = kwargs.get("set_first", True)
        self._first_data = True

        self._data_changed()
    
    def _data_changed(self):
        self.blockSignals(True)
        try:
            data = []
            if self.rois:
                data += self.ivm.rois.keys()
            if self.data:
                data += self.ivm.data.keys()

            current = self.currentText()
            self.clear()
            if self.none_option:
                self.addItem("<none>")

            for name in sorted(data):
                d = self.ivm.data.get(name, self.ivm.rois.get(name, None))
                if d.nvols == 1 or not self.static_only:
                    self.addItem(d.name)

            if self.all_option:
                self.addItem("<all>")

            # Make sure names are visible even with drop down arrow
            width = self.minimumSizeHint().width()
            self.setMinimumWidth(width+50)
        
            idx = self.findText(current)
            if idx >= 0:
                self.setCurrentIndex(idx)
            elif self.none_option:
                self.setCurrentIndex(0)
            elif self.all_option:
                self.setCurrentIndex(len(data))
            else:
                self.setCurrentIndex(-1)
        finally:
            self.blockSignals(False)

        # If requested, initialize when the first data arrives (and send signal)
        if self._set_first and self._first_data and len(data) > 0:
            self.setCurrentIndex(0)
            self._first_data = False

class RoiCombo(OverlayCombo):
    """
    A combo box which gives a choice of ROIs
    """
    def __init__(self, ivm, *args, **kwargs):
        kwargs["rois"] = True
        kwargs["data"] = False
        super(RoiCombo, self).__init__(ivm, *args, **kwargs)
    
class NumericOption(QtGui.QWidget):
    """ Option whose value must be a number (int or float) """

    sig_changed = QtCore.Signal()

    def __init__(self, text, grid, ypos, xpos=0, minval=0, maxval=100, default=0, step=1, decimals=2, intonly=False, spin=True):
        QtGui.QWidget.__init__(self)
        self.use_spin = spin
        self.text = text
        self.minval = minval
        self.maxval = maxval
        self.valid = True

        if intonly:
            self.rtype = int
        else:
            self.rtype = float
            
        self.label = QtGui.QLabel(text)
        grid.addWidget(self.label, ypos, xpos)

        if spin:
            if intonly:
                self.spin = QtGui.QSpinBox()
            else:
                self.spin = QtGui.QDoubleSpinBox()
                self.spin.setDecimals(decimals)
                self.spin.setMinimum(minval)
            self.spin.setMaximum(maxval)
            self.spin.setValue(default)
            self.spin.setSingleStep(step)
            self.spin.valueChanged.connect(self._changed)
            grid.addWidget(self.spin, ypos, xpos+1)
        else:
            self.edit = QtGui.QLineEdit(str(default))
            self.edit.editingFinished.connect(self._edit_changed)
            grid.addWidget(self.edit, ypos, xpos+1)

    def _changed(self):
        self.sig_changed.emit()

    def _edit_changed(self):
        try:
            val = self.rtype(self.edit.text())
            self.valid = (val >= self.minval and val <= self.maxval)
            self.edit.setStyleSheet("")
        except ValueError:
            self.edit.setStyleSheet("QLineEdit {background-color: red}")
            self.valid = False
        self._changed()

    def value(self):
        """ Get the numeric value selected """
        if self.use_spin:
            return self.spin.value()
        elif self.valid:
            return self.rtype(self.edit.text())
        else:
            raise QpException("'%s' is not a valid number")
        
class NumericSlider(QtGui.QWidget):
    """
    Numeric option chooser which uses a slider and two spin boxes
    """
    sig_changed = QtCore.Signal()

    def __init__(self, text, grid, ypos, xpos=0, minval=0, maxval=100, default=0, intonly=False, **kwargs):
        QtGui.QWidget.__init__(self)
        self.text = text
        self.minval = minval
        self.maxval = maxval
        self.hardmin = kwargs.get("hardmin", False)
        self.hardmax = kwargs.get("hardmax", False)
        self.valid = True

        if intonly:
            self.rtype = int
        else:
            self.rtype = float
            
        self.label = QtGui.QLabel(text)
        grid.addWidget(self.label, ypos, xpos)

        hbox = QtGui.QHBoxLayout()
        self.setLayout(hbox)

        #self.min_edit = QtGui.QLineEdit(str(minval))
        #self.min_edit.editingFinished.connect(self_min_changed)
        #hbox.addWidget(self.min_edit)

        self.slider = QtGui.QSlider(QtCore.Qt.Horizontal)
        self.slider.setMaximum(100)
        self.slider.setMinimum(0)
        self.slider.setSliderPosition(int(100 * (default - minval) / (maxval - minval)))
        self.slider.valueChanged.connect(self._slider_changed)
        hbox.addWidget(self.slider)

        self.val_edit = QtGui.QLineEdit(str(default))
        self.val_edit.editingFinished.connect(self._edit_changed)
        hbox.addWidget(self.val_edit)

        grid.addWidget(self, ypos, xpos+1)

    def _changed(self):
        self.sig_changed.emit()

    def _edit_changed(self):
        try:
            val = self.rtype(float(self.val_edit.text()))
            self.valid = True
            self.val_edit.setStyleSheet("")

            if val > self.maxval and not self.hardmax:
                self.maxval = val
            if val < self.minval and not self.hardmin:
                self.minval = val

            val = self.value()
            pos = 100 * (val - self.minval) / (self.maxval - self.minval)
            try:
                self.slider.blockSignals(True)
                self.slider.setSliderPosition(int(pos))
            finally:
                self.slider.blockSignals(False)

        except ValueError:
            self.val_edit.setStyleSheet("QLineEdit {background-color: red}")
            self.valid = False

        self._changed()

    def _slider_changed(self, value):
        val = self.minval + (self.maxval - self.minval) * float(value) / 100
        try:
            self.val_edit.blockSignals(True)
            self.val_edit.setText(sf(val))
        finally:
            self.val_edit.blockSignals(False)
        self._changed()

    def value(self):
        """ Get the numeric value selected """
        if self.valid:
            return self.rtype(float(self.val_edit.text()))
        else:
            raise QpException("'%s' is not a valid number")

    def setLimits(self, minval=None, maxval=None):
        if minval:
            self.minval = minval
        if maxval:
            self.maxval = maxval

        try:
            self.blockSignals(True)
            if self.valid:
                self._edit_changed()
        finally:
            self.blockSignals(False)

class OptionalName(QtGui.QWidget):
    """ String option which can be enabled or disabled """

    sig_changed = QtCore.Signal()

    def __init__(self, text, grid, ypos, xpos=0, default_on=False, default=""):
        QtGui.QWidget.__init__(self)
            
        self.label = QtGui.QCheckBox(text)
        self.label.setChecked(default_on)
        grid.addWidget(self.label, ypos, xpos)

        self.edit = QtGui.QLineEdit(default)
        self.edit.editingFinished.connect(self._edit_changed)
        grid.addWidget(self.edit, ypos, xpos+1)

        self.label.stateChanged.connect(self.edit.setVisible)

    def _changed(self):
        self.sig_changed.emit()

    def _edit_changed(self):
        self._changed()

    def selected(self):
        """ Return True if enabled """
        return self.label.isChecked()

    def value(self):
        """ Return the current text entered """
        return self.edit.text()
        
class ChoiceOption(QtGui.QWidget):
    """ Option which is chosen from a list of possible strings """

    sig_changed = QtCore.Signal()

    def __init__(self, text, grid, ypos, xpos=0, choices=None):
        QtGui.QWidget.__init__(self)
        if choices is None:
            choices = []
        self.choices = choices
        
        self.label = QtGui.QLabel(text)
        grid.addWidget(self.label, ypos, xpos)

        self.combo = QtGui.QComboBox()
        for c in choices:
            self.combo.addItem(c)
        self.combo.currentIndexChanged.connect(self._changed)
        # Bizarre hack to make the dropdown height adjust to the items added
        self.combo.setView(QtGui.QListView())
        grid.addWidget(self.combo, ypos, xpos+1)

    def _changed(self):
        self.sig_changed.emit()

    def value(self):
        """ Get currently selected text """
        return self.combo.currentText()
        
class NumberList(QtGui.QTableWidget):
    """
    Horizontal list of numeric values
    """
    def __init__(self, initial):
        QtGui.QTableWidget.__init__(self, 1, 1)
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setValues(initial)
        self.setFixedHeight(self.rowHeight(0)+5)
        self.default_bg = self.item(0, 0).background()
        self.itemChanged.connect(self._item_changed)
        self.setAcceptDrops(True)

    def valid(self):
        """ Return True if all entries are valid numbers """
        try:
            self.values()
            return True
        except ValueError:
            return False

    def values(self):
        """ Return values as a list of list of floats """
        try:
            return [float(self.item(0, c).text()) for c in range(self.columnCount()-1)]
        except:
            raise RuntimeError("Non-numeric data in list")

    def setValues(self, vals):
        self.blockSignals(True)
        try:
            self.setColumnCount(len(vals)+1)
            for c, val in enumerate(vals):
                self.setItem(0, c, QtGui.QTableWidgetItem("%g" % val))
            self.setItem(0, self.columnCount()-1, QtGui.QTableWidgetItem("..."))
            self.resizeColumnsToContents()
            self.resizeRowsToContents()
        finally:
            self.blockSignals(False)

    def contextMenuEvent(self, event):
        menu = QtGui.QMenu(self)
        insertAction = menu.addAction("Insert before")
        deleteAction = menu.addAction("Delete")
        item = self.itemAt(event.pos())

        action = menu.exec_(self.mapToGlobal(event.pos()))
        if action == insertAction:
            self.insertValue(item.column(), 0)
        elif action == deleteAction:
            self.deleteValue(item.column())
            
    def deleteValue(self, col):
        self.blockSignals(True)
        try:
            for c in range(col, self.columnCount()-1):
                item = self.item(0, c+1)
                self.setItem(0, c, QtGui.QTableWidgetItem(item.text()))
            self.setColumnCount(self.columnCount()-1)
            self.resizeRowsToContents()
        finally:
            self.blockSignals(False)

    def insertValue(self, col, val):
        self.blockSignals(True)
        try:
            self.setColumnCount(self.columnCount()+1)
            for c in range(self.columnCount()-1, col, -1):
                item = self.item(0, c-1)
                self.setItem(0, c, QtGui.QTableWidgetItem(item.text()))
            self.setItem(0, col, QtGui.QTableWidgetItem("%g" % val))
            self.resizeRowsToContents()
        finally:
            self.blockSignals(False)

    def _item_changed(self, item):
        c = item.column()
        if item.text() == "":
            self.deleteValue(c)
        else:
            # Validate new value
            try:
                float(item.text())
                item.setBackground(self.default_bg)
            except ValueError:
                item.setBackground(QtGui.QColor('red'))

            # Add a new column if we have just edited the last one
            if c == self.columnCount() - 1:
                self.blockSignals(True)
                try:
                    self.setColumnCount(self.columnCount()+1)
                    self.setItem(0, self.columnCount()-1, QtGui.QTableWidgetItem("..."))
                finally:
                    self.blockSignals(False)
            self.resizeColumnsToContents()

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
            for url in event.mimeData().urls():
                self.loadFromFile(local_file_from_drop_url(url))
        else:
            event.ignore()

    def loadFromFile(self, filename):
        fvals, nrows, ncols = load_matrix(filename)
        
        if ncols <= 0:
            raise RuntimeError("No numeric data found in file")
        elif ncols == 1:
            self.setValues([r[0] for r in fvals])
        elif nrows == 1:
            self.setValues(fvals[0])
        else:
            # Choose row or column you want
            row, col = self._choose_row_col(fvals)
            if row is not None:
                self.setValues(fvals[row])
            elif col is not None:
                vals = [v[col] for v in fvals]
                self.setValues(vals)

    def _choose_row_col(self, vals):
        d = MatrixViewerDialog(self, vals, title="Choose a row or column", text="Select a row or column containing the data you want")
        if d.exec_():
            ranges = d.table.selectedRanges()
            if ranges:
                r = ranges[0]
                if r.topRow() == r.bottomRow():
                    # Row select
                    return r.topRow(), None
                elif r.leftColumn() == r.rightColumn():
                    # Column select
                    return None, r.leftColumn()
        return None, None    

class LoadNumbers(QtGui.QPushButton):
    """
    PushButton which loads values into a NumberList
    """
    def __init__(self, num_list, label="Load"):
        QtGui.QPushButton.__init__(self, label)
        self.num_list = num_list
        self.clicked.connect(self._button_clicked)

    def _button_clicked(self):
        filename = QtGui.QFileDialog.getOpenFileName()[0]
        if filename:
            self.num_list.loadFromFile(filename)

class NumberGrid(QtGui.QTableView):
    """
    Table of numeric values
    """

    sig_changed = QtCore.Signal()

    def __init__(self, initial, col_headers=None, row_headers=None, expandable=(True, True),
                 fix_height=False, fix_width=False, readonly=False):
        QtGui.QTableView.__init__(self)
        
        self._model = QtGui.QStandardItemModel()
        self.setModel(self._model)
        self.updating = False
        self.expandable = expandable
        self.row_headers = row_headers
        self.col_headers = col_headers
        self.fix_height = fix_height
        self.fix_width = fix_width
        self.default_bg = None
        self.horizontalHeader().hide()
        self.verticalHeader().hide()

        self.setValues(initial, False, col_headers=col_headers, row_headers=row_headers)
        
        if readonly:
            self.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        else:
            self.model().itemChanged.connect(self._item_changed)
        
        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        #self.horizontalHeader().setResizeMode(QtGui.QHeaderView.ResizeToContents)
        self.itemDelegate().closeEditor.connect(self._expand_if_required, QtCore.Qt.QueuedConnection)
        self.setAcceptDrops(True)
        
    def valid(self):
        try:
            self.values()
            return True
        except ValueError:
            return False

    def values(self):
        rows = []
        try:
            for r in range(self._model.rowCount()-int(self.expandable[1])):
                row = [float(self._model.item(r, c).text()) for c in range(self._model.columnCount()-int(self.expandable[0]))]
                rows.append(row)
        except:
            raise ValueError("Non-numeric data in list")
        return rows

    def _set_size(self):
        # QTableWidget is completely incapable of choosing a sensible size. We do our best
        # here but have to allow a bit of random 'padding' in case scrollbars are necessary
        tx, ty = self.horizontalHeader().length()+15, self.verticalHeader().length()+15
        if self.row_headers is not None: tx += self.verticalHeader().width()
        if self.col_headers is not None: ty += self.horizontalHeader().height()
        #sh = self.sizeHint()
        #tx = min(tx, sh.width())
        #ty = min(ty, sh.height())
        if self.fix_height: self.setFixedHeight(ty)
        if self.fix_width: self.setFixedWidth(tx)

    def setValues(self, vals, validate=True, col_headers=None, row_headers=None):
        if validate:
            if vals is None or len(vals) == 0:
                raise ValueError("No values provided")
            elif len(vals[0]) != self._model.columnCount() and not self.expandable[0]:
                raise ValueError("Incorrect number of columns - expected %i" % self._model.columnCount())
            elif len(vals) != self._model.rowCount() and not self.expandable[1]:
                raise ValueError("Incorrect number of rows - expected %i" % self._model.rowCount())

        if (col_headers and self.expandable[0]) or (row_headers and self.expandable[1]):
            raise RuntimeError("Can't specify headers for auto-expandable dimensions")

        self._model.setRowCount(len(vals)+int(self.expandable[1]))
        self._model.setColumnCount(len(vals[0])+int(self.expandable[0]))

        if col_headers:
            self._model.setHorizontalHeaderLabels(col_headers)
            self.horizontalHeader().show()

        if row_headers:
            self._model.setVerticalHeaderLabels(row_headers)
            self.verticalHeader().show()
        
        self._model.blockSignals(True)
        try:
            for r, rvals in enumerate(vals):
                for c, v in enumerate(rvals):
                    item = QtGui.QStandardItem("%g" % v)
                    self._model.setItem(r, c, item)
            
            if self.expandable[0]:
                for r in range(len(vals)):
                    item = QtGui.QStandardItem("")
                    self._model.setItem(r, self._model.columnCount()-1, item)
        
            if self.expandable[1]:
                for c in range(len(vals[0])):
                    item = QtGui.QStandardItem("")
                    self._model.setItem(self._model.rowCount()-1, c, item)
        
            self.resizeColumnsToContents()
            self.resizeRowsToContents()
        finally:
            self._model.blockSignals(False)
            
        self._set_size()
        self.sig_changed.emit()

    def _range_empty(self, cs, rs):
        empty = True
        for r in rs:
            for c in cs:
                # Check if r or c is negative and do not consider empty
                # so last row/column will not be deleted
                if r < 0 or c < 0 or (self._model.item(r, c) is not None and self._model.item(r, c).text() != ""):
                    empty = False
                    break
        return empty

    def _item_changed(self, item):
        if self.updating: return
        self.updating = True
        try:
            if self.default_bg is None:
                self.default_bg = item.background()

            try:
                float(item.text())
                item.setBackground(self.default_bg)
            except ValueError:
                if item.text() != "": 
                    item.setBackground(QtGui.QColor('red'))

            self.resizeColumnsToContents()
            self.resizeRowsToContents()
        finally:
            self.updating = False

    def _expand_if_required(self):
        last_col = self._model.columnCount()-1
        last_row = self._model.rowCount()-1
        if self.expandable[0]:
            if not self._range_empty([last_col], range(self._model.rowCount())):
                # Last column is not empty - add a new empty column
                self._model.setColumnCount(self._model.columnCount()+1)
                # Set rest of new column to prevent it from immediately being invalid
                for row in range(self._model.rowCount()-int(self.expandable[1])):
                    item = self._model.item(row, last_col)
                    if item is None or item.text() == "":
                        item = QtGui.QStandardItem(self._model.item(row, last_col-1).text())
                        self._model.setItem(row, last_col, item)
                    
            elif self._range_empty([self._model.columnCount()-2], range(self._model.rowCount())):
                # Last two columns are empty, so remove last column
                self._model.setColumnCount(self._model.columnCount()-1)

        if self.expandable[1]:
            if not self._range_empty(range(self._model.columnCount()), [last_row]):
               # Last row is not empty - add a new empty row
                self._model.setRowCount(self._model.rowCount()+1)
                # Set rest of new row to prevent it from immediately being invalid
                for col in range(self._model.columnCount()-int(self.expandable[0])):
                    item = self._model.item(last_row, col)
                    if item is None or item.text() == "":
                        self._model.setItem(last_row, col, QtGui.QStandardItem(self._model.item(last_row-1, col).text()))
            
            elif self._range_empty(range(self._model.columnCount()), [self._model.rowCount()-2]):
                # Last-but-one row is empty, so remove last row (which is always empty)
                self._model.setColumnCount(self._model.columnCount()-1)
        self.sig_changed.emit()

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
            for url in event.mimeData().urls():
                self.loadFromFile(local_file_from_drop_url(url))
        else:
            event.ignore()
            
    def loadFromFile(self, filename):
        fvals, _, ncols = load_matrix(filename)
        if ncols <= 0:
            raise RuntimeError("No numeric data found in file")
        else:
            self.setValues(fvals)

class NumberVList(NumberGrid):
    """
    Single-column NumberGrid
    """
    def __init__(self, initial, headers=None, expandable=True, fix_height=False):
        NumberGrid.__init__(self, initial, row_headers=headers, 
                            expandable=(False, expandable), fix_height=fix_height)

    def values(self):
        return [v[0] for v in NumberGrid.values(self)]

    def setValues(self, values, validate=True):
        NumberGrid.setValues(self, [[v,] for v in values], validate)

    def loadFromFile(self, filename):
        fvals, _, ncols = load_matrix(filename)
        
        if ncols <= 0:
            raise RuntimeError("No numeric data found in file")
        elif ncols == 1:
            self.setValues(fvals[0])
        else:
            # Choose row or column you want
            row, col = self._choose_row_col(fvals)
            if row is not None:
                self.setValues(fvals[row])
            elif col is not None:
                vals = [v[col] for v in fvals]
                self.setValues(vals)

    def _choose_row_col(self, vals):
        d = MatrixViewerDialog(self, vals, title="Choose a row or column", text="Select a row or column containing the data you want")
        if d.exec_():
            ranges = d.table.selectedRanges()
            if len(ranges) > 0:
                r = ranges[0]
                if r.topRow() == r.bottomRow():
                    # Row select
                    return r.topRow(), None
                elif r.leftColumn() == r.rightColumn():
                    # Column select
                    return None, r.leftColumn()
        return None, None    

class NumberHList(NumberVList):
    """
    Single-row NumberGrid
    """
    def __init__(self, initial, headers=None, expandable=True, fix_width=False):
        NumberGrid.__init__(self, initial, col_headers=headers, 
                            expandable=(expandable, False), fix_width=fix_width)

    def values(self):
        return NumberGrid.values(self)[0]

    def setValues(self, values, validate=True):
        NumberGrid.setValues(self, [values,], validate)

class Citation(QtGui.QWidget):
    def __init__(self, title, author, journal):
        QtGui.QWidget.__init__(self)
        self.title, self.author, self.journal = title, author, journal
        hbox = QtGui.QHBoxLayout()
        self.setLayout(hbox)

        btn = QtGui.QPushButton()
        icon = QtGui.QIcon(get_icon("citation"))
        btn.setIcon(icon)
        btn.setSizePolicy(QtGui.QSizePolicy.Maximum, QtGui.QSizePolicy.Maximum)
        btn.clicked.connect(self.lookup)
        hbox.addWidget(btn)
        hbox.setAlignment(btn, QtCore.Qt.AlignTop)

        text = "<font size=3><i>" + title + "</i><br>" + author + "<br>" + journal + "</font>"
        label = QtGui.QLabel(text)
        label.setWordWrap(True)
        hbox.addWidget(label)
        
    def lookup(self):
        import webbrowser, urllib
        search_terms = tuple([urllib.quote(s) for s in (self.title, self.author, self.journal)])
        url = "https://www.google.com/search?q=%s+%s+%s" % search_terms
        webbrowser.open(url, new=0, autoraise=True)

class OptionsButton(QtGui.QPushButton):
    def __init__(self, widget=None):
        QtGui.QPushButton.__init__(self)
        self.setIcon(QtGui.QIcon(get_icon("options.png")))
        self.setIconSize(QtCore.QSize(14, 14))
        if widget:
            self.clicked.connect(widget.show_options)

class TitleWidget(QtGui.QWidget):
    def __init__(self, widget, title=None, subtitle=None, help="", help_btn=True, batch_btn=True, opts_btn=False, icon=True):
        QtGui.QWidget.__init__(self)
        if title is None:
            title = widget.name
        if subtitle is None:
            subtitle = widget.description
            
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)         
        hbox = QtGui.QHBoxLayout()
        if icon and hasattr(widget, "icon"):
            icon = QtGui.QLabel()
            icon.setPixmap(widget.icon.pixmap(32, 32))
            hbox.addWidget(icon)
        hbox.addWidget(QtGui.QLabel('<font size="5">%s</font>' % title))   
        hbox.addStretch(1)
        if batch_btn: hbox.addWidget(BatchButton(widget))
        if help_btn: hbox.addWidget(HelpButton(self, help))
        if opts_btn: hbox.addWidget(OptionsButton(widget))
        vbox.addLayout(hbox)

        hbox = QtGui.QHBoxLayout()
        hbox.addWidget(QtGui.QLabel(subtitle))      
        hbox.addStretch(1)
        hbox.addWidget(QtGui.QLabel(widget.version))
        vbox.addLayout(hbox)

class RunBox(QtGui.QGroupBox, LogSource):

    sig_postrun = QtCore.Signal()

    """
    Box containing a 'run' button, a progress bar, a 'cancel' button and a 'view log' button 

    Designed for use with BackgroundTask
    """
    def __init__(self, get_process_fn=None, get_rundata_fn=None, widget=None, ivm=None, title="Run", btn_label="Run", save_option=False):
        LogSource.__init__(self)
        QtGui.QGroupBox.__init__(self)
        self.save_option = save_option
        
        self.setTitle(title)
        self.setSizePolicy(QtGui.QSizePolicy.MinimumExpanding, QtGui.QSizePolicy.MinimumExpanding)
        
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        hbox = QtGui.QHBoxLayout()
        self.runBtn = QtGui.QPushButton(btn_label, self)
        self.runBtn.clicked.connect(self.start)
        hbox.addWidget(self.runBtn)
        self.progress = QtGui.QProgressBar(self)
        hbox.addWidget(self.progress)
        self.cancelBtn = QtGui.QPushButton('Cancel', self)
        self.cancelBtn.clicked.connect(self._cancel)
        self.cancelBtn.setEnabled(False)
        hbox.addWidget(self.cancelBtn)
        self.logBtn = QtGui.QPushButton('View log', self)
        self.logBtn.clicked.connect(self._view_log)
        self.logBtn.setEnabled(False)
        hbox.addWidget(self.logBtn)
        vbox.addLayout(hbox)

        self.step_label = QtGui.QLabel()
        self.step_label.setVisible(False)
        vbox.addWidget(self.step_label) 

        if self.save_option:
            hbox = QtGui.QHBoxLayout()
            self.save_cb = QtGui.QCheckBox("Save copy of output data")
            hbox.addWidget(self.save_cb)
            self.save_folder_edit = QtGui.QLineEdit()
            hbox.addWidget(self.save_folder_edit)
            btn = QtGui.QPushButton("Choose folder")
            btn.clicked.connect(self._choose_output_folder)
            hbox.addWidget(btn)
            self.save_cb.stateChanged.connect(self.save_folder_edit.setEnabled)
            self.save_cb.stateChanged.connect(btn.setEnabled)
            vbox.addLayout(hbox)

        self.get_process_fn = get_process_fn
        self.get_rundata_fn = get_rundata_fn
        self.widget = widget
        self.ivm = ivm
        self.process = None

    def start(self):
        """
        Start running the process
        """
        if self.get_process_fn is not None:
            self.process = self.get_process_fn()
            rundata = self.get_rundata_fn()
        else:
            processes = self.widget.processes()
            if isinstance(processes, dict):
                processes = [processes,]
            self.process = Script(self.ivm, error_action=Script.FAIL, embed_log=True)
            rundata = {"parsed-yaml" : {"Processing" : processes}}

        self.progress.setValue(0)
        self.runBtn.setEnabled(False)
        self.cancelBtn.setEnabled(True)
        self.logBtn.setEnabled(False)
        self.step_label.setVisible(False)

        self.process.sig_finished.connect(self._finished)
        self.process.sig_progress.connect(self._update_progress)
        self.process.sig_step.connect(self._new_step)
        self.process.execute(rundata)

    def _cancel(self):
        self.process.cancel()

    def _update_progress(self, complete):
        self.progress.setValue(100*complete)

    def _new_step(self, desc):
        self.step_label.setText(desc)
        self.step_label.setVisible(True)

    def _finished(self, status, log, exception):
        try:
            self.debug("RunBox: Finished: %i %i %s", status, len(log), exception)
            self.log = log
            if status == Process.SUCCEEDED:
                self.progress.setValue(100)
                self.step_label.setVisible(False)
                if self.save_option and self.save_cb.isChecked() and self.save_folder_edit.text():
                    self.process.save_output(self.save_folder_edit.text())
            elif status == Process.CANCELLED:
                self.progress.setValue(0)
                self.step_label.setVisible(False)
            elif isinstance(exception, BaseException):
                if self.debug_enabled(): 
                    traceback.print_exc(exception)
                raise exception
            else:
                raise QpException("Process finished with error status %i but no error was returned" % status)
        finally:
            if self.process is not None:
                self.process.sig_finished.disconnect(self._finished)
                self.process.sig_progress.disconnect(self._update_progress)
                self.process.sig_step.disconnect(self._new_step)
                self.process = None
            self.runBtn.setEnabled(True)
            self.logBtn.setEnabled(True)
            self.cancelBtn.setEnabled(False)
            self.sig_postrun.emit()
            
    def _view_log(self):
        self.logview = TextViewerDialog(text=self.log, parent=self)
        self.logview.show()
        self.logview.raise_()

    def _choose_output_folder(self):
        outputDir = QtGui.QFileDialog.getExistingDirectory(self, 'Choose directory to save output')
        if outputDir:
            self.save_folder_edit.setText(outputDir)

class RunButton(QtGui.QPushButton, LogSource):
    """
    Simple button to run the processing associated with a QpWidget

    Designed for use with QpWidget that implements the ``processes`` method. 
    """

    sig_postrun = QtCore.Signal()

    def __init__(self, widget, label="Run"):
        LogSource.__init__(self)
        QtGui.QPushButton.__init__(self, label)

        self.clicked.connect(self._start)
        self.widget = widget
        self.process = Script(widget.ivm, error_action=Script.FAIL, embed_log=True)
        self.process.sig_finished.connect(self._finished)

    def _start(self):
        # FIXME spinner
        processes = self.widget.processes()
        if isinstance(processes, dict):
            processes = [processes,]
        self.process.execute({"parsed-yaml" : {"Processing" : processes}})

    def _finished(self, status, log, exception):
        try:
            self.debug("RunButton: Finished: %i %i %s", status, len(log), exception)
            if isinstance(exception, BaseException):
                if self.debug_enabled(): 
                    traceback.print_exc(exception)
                raise exception
            elif status != Process.SUCCEEDED:
                raise QpException("Process finished with error status %i but no error was returned" % status)
        finally:
            self.sig_postrun.emit()

class RunWidget(QtGui.QGroupBox, LogSource):
    """
    Box containing a 'run' button, a progress bar, a 'cancel' button and a 'view log' button

    Designed for use with QpWidget that implements the ``processes`` method. 
    """

    sig_postrun = QtCore.Signal()

    def __init__(self, widget, title="Run", btn_label="Run", save_option=False):
        LogSource.__init__(self)
        QtGui.QGroupBox.__init__(self)
        self.save_option = save_option
        
        self.setTitle(title)
        self.setSizePolicy(QtGui.QSizePolicy.MinimumExpanding, QtGui.QSizePolicy.MinimumExpanding)
        
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        hbox = QtGui.QHBoxLayout()
        self.runBtn = QtGui.QPushButton(btn_label, self)
        self.runBtn.clicked.connect(self.start)
        hbox.addWidget(self.runBtn)
        self.progress = QtGui.QProgressBar(self)
        hbox.addWidget(self.progress)
        self.cancelBtn = QtGui.QPushButton('Cancel', self)
        self.cancelBtn.clicked.connect(self._cancel)
        self.cancelBtn.setEnabled(False)
        hbox.addWidget(self.cancelBtn)
        self.logBtn = QtGui.QPushButton('View log', self)
        self.logBtn.clicked.connect(self._view_log)
        hbox.addWidget(self.logBtn)
        vbox.addLayout(hbox)

        self.step_label = QtGui.QLabel()
        self.step_label.setVisible(False)
        vbox.addWidget(self.step_label) 

        if self.save_option:
            hbox = QtGui.QHBoxLayout()
            self.save_cb = QtGui.QCheckBox("Save copy of output data")
            hbox.addWidget(self.save_cb)
            self.save_folder_edit = QtGui.QLineEdit()
            hbox.addWidget(self.save_folder_edit)
            btn = QtGui.QPushButton("Choose folder")
            btn.clicked.connect(self._choose_output_folder)
            hbox.addWidget(btn)
            self.save_cb.stateChanged.connect(self.save_folder_edit.setEnabled)
            self.save_cb.stateChanged.connect(btn.setEnabled)
            vbox.addLayout(hbox)

        self.widget = widget
        self.logview = TextViewerDialog(parent=self)
        self.process = Script(widget.ivm, error_action=Script.FAIL, embed_log=True)

        self.process.sig_finished.connect(self._finished)
        self.process.sig_progress.connect(self._progress)
        self.process.sig_step.connect(self._step)
        self.process.sig_log.connect(self._log)

    def start(self):
        self.log = ""
        self.logview.text = self.log
        self.progress.setValue(0)
        self.runBtn.setEnabled(False)
        self.cancelBtn.setEnabled(True)
        self.step_label.setVisible(False)

        processes = self.widget.processes()
        if isinstance(processes, dict):
            processes = [processes,]
        self.process.execute({"parsed-yaml" : {"Processing" : processes}})

    def _cancel(self):
        self.process.cancel()

    def _progress(self, complete):
        self.progress.setValue(100*complete)

    def _step(self, desc):
        self.step_label.setText(desc)
        self.step_label.setVisible(True)

    def _log(self, msg):
        self.logview.text = self.process.get_log()

    def _finished(self, status, log, exception):
        try:
            self.debug("RunWidget: Finished: %i %i %s", status, len(log), exception)
            self.log = log
            if status == Process.SUCCEEDED:
                self.progress.setValue(100)
                self.step_label.setVisible(False)
                if self.save_option and self.save_cb.isChecked() and self.save_folder_edit.text():
                    self.process.save_output(self.save_folder_edit.text())
            elif status == Process.CANCELLED:
                self.progress.setValue(0)
                self.step_label.setVisible(False)
            elif isinstance(exception, BaseException):
                if self.debug_enabled(): 
                    traceback.print_exc(exception)
                raise exception
            else:
                raise QpException("Process finished with error status %i but no error was returned" % status)
        finally:
            self.runBtn.setEnabled(True)
            self.cancelBtn.setEnabled(False)
            self.sig_postrun.emit()
            
    def _view_log(self):
        self.logview.show()
        self.logview.raise_()

    def _choose_output_folder(self):
        outputDir = QtGui.QFileDialog.getExistingDirectory(self, 'Choose directory to save output')
        if outputDir:
            self.save_folder_edit.setText(outputDir)

class OrderList(QtGui.QListView):
    """
    Vertical list of items which can be re-ordered but not changed directly
    """
    
    sig_changed = QtCore.Signal()

    def __init__(self, initial=(), col_headers=None):
        QtGui.QListView.__init__(self)
        self.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        #self.setSizePolicy(QtGui.QSizePolicy.Preferred, QtGui.QSizePolicy.Preferred)
        self.model = QtGui.QStandardItemModel()
        self.setModel(self.model)
        #if col_headers:
        #    self.setVerticalHeaderLabels(col_headers)
        #else:
        #    self.horizontalHeader().hide()
        self.setItems(initial)

    def setItems(self, items):
        self.blockSignals(True)
        try:
            self.model.clear()
            for r, item in enumerate(items):
                item = QtGui.QStandardItem(item)
                self.model.appendRow(item)
            height = 0
            for r in range(len(items)):
                height += self.rectForIndex(self.model.index(r, 0)).height()
            self.setFixedHeight(height + 2*len(items))
        finally:
            self.blockSignals(False)
            self.sig_changed.emit()

    def items(self):
        return [self.model.item(r).text() for r in range(self.model.rowCount())]

    def currentUp(self):
        """ Move currently selected item up"""
        idx = self.currentIndex().row()
        if idx > 0:
            items = self.items()
            temp = items[idx-1]
            items[idx-1] = items[idx]
            items[idx] = temp
            self.setItems(items)
            self.setCurrentIndex(self.model.index(idx-1, 0))
            self.sig_changed.emit()

    def currentDown(self):
        """ Move currently selected item down"""
        idx = self.currentIndex().row()
        if idx < self.model.rowCount() - 1:
            items = self.items()
            temp = items[idx+1]
            items[idx+1] = items[idx]
            items[idx] = temp
            self.setItems(items)
            self.setCurrentIndex(self.model.index(idx+1, 0))
            self.sig_changed.emit()

class OrderListButtons(QtGui.QVBoxLayout):
    def __init__(self, orderlist):
        QtGui.QVBoxLayout.__init__(self)
        self.list = orderlist
        self.up_btn = QtGui.QPushButton()
        self.up_btn.setIcon(QtGui.QIcon(get_icon("up.png")))
        self.up_btn.setFixedSize(16, 16)
        self.up_btn.clicked.connect(self.list.currentUp)
        self.addWidget(self.up_btn)
        self.down_btn = QtGui.QPushButton()
        self.down_btn.setIcon(QtGui.QIcon(get_icon("down.png")))
        self.down_btn.setFixedSize(16, 16)
        self.down_btn.clicked.connect(self.list.currentDown)
        self.addWidget(self.down_btn)

class WarningBox(QtGui.QFrame):
    """
    Widget which just displays a warning, e.g. when a QpWidget can't be used for some reason
    """

    def __init__(self, text=""):
        QtGui.QFrame.__init__(self)
        hbox = QtGui.QHBoxLayout()
        hbox.setSpacing(0)
        hbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(hbox)

        #self.warn_icon = QtGui.QIcon.fromTheme("dialog-error")
        self.icon = QtGui.QLabel()
        self.warn_icon = self.icon.style().standardIcon(QtGui.QStyle.SP_MessageBoxWarning)
        self.icon.setPixmap(self.warn_icon.pixmap(32, 32))
        hbox.addWidget(self.icon)

        self.text = QtGui.QLabel()
        self.text.setWordWrap(True)
        hbox.addWidget(self.text)
        hbox.setStretchFactor(self.text, 2)

        self.setStyleSheet("QWidget { background-color: #ddcca8; color: black; padding: 5px 5px 5px 5px; border-radius: 5;}")
        self.warn(text)

    def warn(self, text=""):
        if text:
            self.text.setText(text)
            self.setVisible(True)
        else:
            self.clear()
        
    def clear(self):
        self.setVisible(False)

class MultiExpander(QtGui.QWidget):
    """
    Generic expander widget, alternative to tab box which allows all to be 'closed'
    """
    def __init__(self, widgets, parent=None, default_visible=None):
        super(MultiExpander, self).__init__(parent)

        vbox = QtGui.QVBoxLayout()

        self.arrow_right = self.style().standardIcon(QtGui.QStyle.SP_ArrowRight)
        self.arrow_down = self.style().standardIcon(QtGui.QStyle.SP_ArrowDown)
        self.widgets = widgets
        self.toggle_btns = {}

        btn_hbox = QtGui.QHBoxLayout()
        w_hbox = QtGui.QHBoxLayout()
        for name, w in self.widgets.items():
            if name == default_visible:
                w.setVisible(True)
                self.toggle_btns[name] = QtGui.QPushButton(self.arrow_down, name)
            else:
                w.setVisible(False)
                self.toggle_btns[name] = QtGui.QPushButton(self.arrow_right, name)
            self.toggle_btns[name].clicked.connect(self._toggle(name))
            btn_hbox.addWidget(self.toggle_btns[name])
            w_hbox.addWidget(w)

        vbox.addLayout(btn_hbox)
        vbox.addLayout(w_hbox)
        self.setLayout(vbox)

    def _toggle(self, name):
        """ Return 'toggle' callback for named widget """
        def _cb():
            for wname, w in self.widgets.items():
                if wname == name and not w.isVisible():
                    self.toggle_btns[wname].setIcon(self.arrow_down)
                    w.setVisible(True)
                else:
                    self.toggle_btns[wname].setIcon(self.arrow_right)
                    w.setVisible(False)
        return _cb

class FslDirWidget(QtGui.QFrame):
    """
    Widget which reports current FSLDIR and allows it to be changed
    """
    sig_changed = QtCore.Signal(str)

    def __init__(self, **kwargs):
        QtGui.QFrame.__init__(self, **kwargs)
        self.setStyleSheet("QWidget { background-color: #609050; color: black; border-radius: 5;}")
        self._settings = QtCore.QSettings()

        hbox = QtGui.QHBoxLayout()
        self.setLayout(hbox)
        icon = QtGui.QLabel()
        info_icon = icon.style().standardIcon(QtGui.QStyle.SP_MessageBoxInformation)
        icon.setPixmap(info_icon.pixmap(32, 32))
        hbox.addWidget(icon)
        lbl = QtGui.QLabel("Using FSL in")
        hbox.addWidget(lbl)
        hbox.setAlignment(lbl, QtCore.Qt.AlignTop) # Because the elided label always goes to the top...
        vbox = QtGui.QVBoxLayout() 
        self._fsldir_label = ElidedLabel()
        vbox.addWidget(self._fsldir_label)
        self._fsldevdir_label = ElidedLabel()
        vbox.addWidget(self._fsldevdir_label)
        hbox.addLayout(vbox)

        btn = QtGui.QPushButton("Change")
        btn.clicked.connect(self._change_fsldir)
        hbox.addWidget(btn)
        self._update_label()

    @property
    def fsldir(self):
        """
        :return: Location of FSL installation

        This looks for a previously configured location from Quantiphyse, or alternatively
        for the FSLDIR environment variable, and lastly will check a few common
        locations.

        If a location is found which is not stored in the FSLDIR environment variable then
        this variable will be updated.
        """
        fsldir, _ = self._get_fsl_dirs()
        return fsldir

    @property
    def fsldevdir(self):
        """
        :return: Location of updated FSL code

        This looks for a previously configured location from Quantiphyse, or alternatively
        for the FSLDEVDIR environment variable.

        If a location is found which is not stored in the FSLDEVDIR environment variable then
        this variable will be updated.
        """
        _, fsldevdir = self._get_fsl_dirs()
        return fsldevdir

    def _get_fsl_dirs(self):
        if self._settings.contains("fslqp/fsldir"):
            os.environ["FSLDIR"] = self._settings.value("fslqp/fsldir")
        elif "FSLDIR" not in os.environ:
            places_to_try = [
                "/usr/local/fsl",
                "/opt/fsl",
            ]
            for place in places_to_try:
                if self._possible_fsldir(place):
                    os.environ["FSLDIR"] = place
                    break

        if self._settings.contains("fslqp/fsldevdir"):
            os.environ["FSLDEVDIR"] = self._settings.value("fslqp/fsldevdir")

        return os.environ.get("FSLDIR", None), os.environ.get("FSLDEVDIR", None)

    def _change_fsldir(self):
        changed = False
        dialog = FslDirDialog(self.fsldir, self.fsldevdir)
        response = dialog.exec_()
        if response:
            if self._possible_fsldir(dialog.fsldir):
                self._settings.setValue("fslqp/fsldir", dialog.fsldir)
            else:
                raise QpException("Selected directory does not appear to contain FSL")

            if dialog.fsldevdir:
                self._settings.setValue("fslqp/fsldevdir", dialog.fsldevdir)
            else:
                self._settings.setValue("fslqp/fsldevdir", "")
            
            changed = True
           
        if changed:
            self._update_label()
            self.sig_changed.emit(self.fsldir)
            
    def _update_label(self):
        text = []
        if self.fsldir:
            text.append(self.fsldir)
        if self.fsldevdir:
            text.append(self.fsldevdir)

        self._fsldevdir_label.setText("")
        if len(text) > 0:
            self._fsldir_label.setText(text[0])
            self._fsldir_label.setToolTip(text[0])
            if len(text) > 1:
                self._fsldevdir_label.setText(text[1])
                self._fsldevdir_label.setToolTip(text[1])
        else:
            self._fsldir_label.setText("FSLDIR not set - click button to set")
            self._fsldir_label.setToolTip("")

    def _possible_fsldir(self, folder):
        return os.path.exists(os.path.join(folder, "bin", "flirt"))

class FslDirDialog(QtGui.QDialog):
    """
    Dialog box to choose FSLDIR
    """

    def __init__(self, fsldir, fsldevdir):
        super(FslDirDialog, self).__init__(quantiphyse.gui.dialogs.MAINWIN)
        self.setWindowTitle("Choose location of FSL installation")
        vbox = QtGui.QVBoxLayout()

        self.optbox = OptionBox()
        self.optbox.add("FSL installation", FileOption(dirs=True, initial=fsldir), key="fsldir")
        self.optbox.add("FSL development code", FileOption(dirs=True, initial=fsldevdir), checked=True, enabled=bool(fsldevdir), key="fsldevdir")
        vbox.addWidget(self.optbox)
        
        hbox = QtGui.QHBoxLayout()
        hbox.addStretch(1)
        self.button_box = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        hbox.addWidget(self.button_box)
        vbox.addLayout(hbox)
        self.setLayout(vbox)

    @property
    def fsldir(self):
        return self.optbox.values()["fsldir"]
        
    @property
    def fsldevdir(self):
        return self.optbox.values().get("fsldevdir", None)

class ElidedLabel(QtGui.QFrame):
    """
    Equivalent to a QLabel but uses ellipsis to clip long text and prevents the
    label from growing beyond it's natural size
    
    Converted to Python from C++ example code::

        https://stackoverflow.com/questions/7381100/text-overflow-for-a-qlabel-s-text-rendering-in-qt
    """
    def __init__(self, text="", parent=None):
        QtGui.QFrame.__init__(self, parent)
        self._content = text
        self._elided = False
        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred)

    def setText(self, text):
        self._content = text
        self.update()
    
    def paintEvent(self, event):
        QtGui.QFrame.paintEvent(self, event)
        painter = QtGui.QPainter(self)
        metrics = painter.fontMetrics()
        line_spacing = metrics.lineSpacing()
        y = 0
        elided = False

        text_layout = QtGui.QTextLayout(self._content, painter.font())
        text_layout.beginLayout()
        while 1:
            line = text_layout.createLine()

            if not line.isValid():
                break

            line.setLineWidth(self.width())
            nextLineY = y + line_spacing

            if self.height() >= nextLineY + line_spacing:
                line.draw(painter, QtCore.QPoint(0, y))
                y = nextLineY
            else:
                lastLine = self._content[line.textStart():]
                elidedLastLine = metrics.elidedText(lastLine, QtCore.Qt.ElideRight, self.width())
                painter.drawText(QtCore.QPoint(0, y + metrics.ascent()), elidedLastLine)
                line = text_layout.createLine()
                elided = line.isValid()
                break
        text_layout.endLayout()
        self._elided = elided



import unittest
import time

import numpy as np

from quantiphyse.processes import Process
from quantiphyse.test.widget_test import WidgetTest

from .widget import BatchBuilderWidget

CLUSTER = """
  - KMeans:
      data: data_3d
      roi: mask
      n-clusters: 4
      output-name: clusters
"""
class BatchBuilderWidgetTest(WidgetTest):

    def widget_class(self):
        return BatchBuilderWidget

    def testNoData(self):
        self.assertTrue(self.w.proc_edit.toPlainText() == "")
        if self.w.run_box.runBtn.isEnabled():
            self.w.run_box.runBtn.clicked.emit()
        self.processEvents()
        self.assertFalse(self.error)

    def testDataLoaded(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.processEvents()
        self.assertTrue(self.w.proc_edit.toPlainText() != "")

        self.w.run_box.runBtn.clicked.emit()
        self.processEvents()
        self.assertFalse(self.error)

    def testAddProcess(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.ivm.add(self.mask, grid=self.grid, name="mask")
        self.processEvents()
        yaml = self.w.proc_edit.toPlainText()
        add_str = "# Additional processing steps go here\n"
        add_idx = yaml.find(add_str)
        self.assertTrue(add_idx > 0)
        add_idx += len(add_str)
        pre = yaml[:add_idx]
        post = yaml[add_idx:]
        self.w.proc_edit.setPlainText(pre + CLUSTER + post)
        self.processEvents()
        
        self.w.run_box.runBtn.clicked.emit()
        while not self.error and not hasattr(self.w.run_box, "log"):
            self.processEvents()
            time.sleep(2)
            
        self.assertEqual(self.w.process.status, Process.SUCCEEDED)
        self.assertFalse(self.error)
        self.assertTrue("clusters" in self.ivm.rois)

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Subclasses of QpData for loading and saving data in different formats

Copyright (c) 2013-2018 University of Oxford
"""
from __future__ import division, print_function

import sys
import os
import warnings
import glob
import logging

import nibabel as nib
import numpy as np

from quantiphyse.utils import QpException
from .qpdata import DataGrid, QpData, NumpyData
from .nifti import NiftiData, save as save_nifti
from .dicoms import DicomFolder

LOG = logging.getLogger(__name__)

def load(fname):
    """
    Load a data file

    :return: QpData instance
    """
    if os.path.isdir(fname):
        return DicomFolder(fname)
    elif fname.endswith(".nii") or fname.endswith(".nii.gz"):
        return NiftiData(fname)
    else:
        raise QpException("%s: Unrecognized file type" % fname)

def save(data, fname, grid=None, outdir=""):
    """
    Save data to a file
    
    :param data: QpData instance
    :param fname: File name
    :param grid: If specified, grid to save the data on
    :param outdir: Optional output directory if fname is not absolute
    """
    save_nifti(data, fname, grid, outdir)



import unittest
import re

import numpy as np

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import DataGrid
from quantiphyse.test.widget_test import WidgetTest

from .widgets import DataStatistics, MultiVoxelAnalysis, VoxelAnalysis, MeasureWidget

class DataStatisticsTest(WidgetTest):

    def widget_class(self):
        return DataStatistics

    def testNoData(self):
        """ User clicks the show buttons with no data"""
        self.harmless_click(self.w.butgen)
        self.harmless_click(self.w.butgenss)

    def test3dData(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.w.data.value = ["data_3d",]
        self.harmless_click(self.w.butgen)
        self.assertTrue(self.w.stats_table.isVisible())
        model = self.w.stats_table.model()
        self.assertEquals(model.rowCount(), 5)
        self.assertEquals(model.columnCount(), 1)
        self.assertAlmostEquals(float(model.item(0, 0).text()), np.mean(self.data_3d), delta=0.01)
        self.assertAlmostEquals(float(model.item(1, 0).text()), np.median(self.data_3d), delta=0.01)
        self.assertAlmostEquals(float(model.item(2, 0).text()), np.std(self.data_3d), delta=0.01)
        self.assertAlmostEquals(float(model.item(3, 0).text()), np.min(self.data_3d), delta=0.01)
        self.assertAlmostEquals(float(model.item(4, 0).text()), np.max(self.data_3d), delta=0.01)
        
    def testAllData(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        # Select 'all data'
        self.w.data.value = ["data_3d", "data_4d"]
        self.harmless_click(self.w.butgen)
        self.assertTrue(self.w.stats_table.isVisible())
        self.assertEquals(self.w.stats_table.model().rowCount(), 5)
        self.assertEquals(self.w.stats_table.model().columnCount(), 2)

    def testShowHide(self):
        self.harmless_click(self.w.butgen)
        self.assertTrue(self.w.stats_table.isVisible())
        self.harmless_click(self.w.butgen)
        self.assertFalse(self.w.stats_table.isVisible())
     
    def testCopy(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.w.data.value = ["data_3d",]
        self.harmless_click(self.w.butgen)
        self.harmless_click(self.w.copy_btn)
        cb = QtGui.QApplication.clipboard()
        rows = [row.split("\t") for row in cb.text().split("\n")]
        self.assertEquals(rows[0][0], "")
        self.assertEquals(rows[0][1].strip(), "data_3d")
        numbers = [float(row[1]) for row in rows[1:] if len(row) > 1]
        self.assertTrue(len(numbers) == 5)
        
    def testSliceStats(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.w.data.value = ["data_3d",]
        self.harmless_click(self.w.butgenss)
        self.ivl.set_focus([0, 0, 2, 0])
        model = self.w.stats_table_ss.model()
        self.assertEquals(model.rowCount(), 5)
        self.assertEquals(model.columnCount(), 1)

        self.w.sscombo.setCurrentIndex(0) # Axial slice (I/S)
        self.processEvents()
        self.assertEquals(self.w.sscombo.currentText(), "Axial")
        self.assertAlmostEquals(float(model.item(0, 0).text()), np.mean(self.data_3d[:,:,2]), delta=0.01)
        self.assertAlmostEquals(float(model.item(1, 0).text()), np.median(self.data_3d[:,:,2]), delta=0.01)
        self.assertAlmostEquals(float(model.item(2, 0).text()), np.std(self.data_3d[:,:,2]), delta=0.01)
        self.assertAlmostEquals(float(model.item(3, 0).text()), np.min(self.data_3d[:,:,2]), delta=0.01)
        self.assertAlmostEquals(float(model.item(4, 0).text()), np.max(self.data_3d[:,:,2]), delta=0.01)
          
        self.w.sscombo.setCurrentIndex(1) # Coronal slice (A/P)
        self.ivl.set_focus([0, 4, 0, 0])
        self.processEvents()
        self.assertEquals(self.w.sscombo.currentText(), "Coronal")
        self.assertAlmostEquals(float(model.item(0, 0).text()), np.mean(self.data_3d[:,4,:]), delta=0.01)
        self.assertAlmostEquals(float(model.item(1, 0).text()), np.median(self.data_3d[:,4,:]), delta=0.01)
        self.assertAlmostEquals(float(model.item(2, 0).text()), np.std(self.data_3d[:,4,:]), delta=0.01)
        self.assertAlmostEquals(float(model.item(3, 0).text()), np.min(self.data_3d[:,4,:]), delta=0.01)
        self.assertAlmostEquals(float(model.item(4, 0).text()), np.max(self.data_3d[:,4,:]), delta=0.01)
        
        self.w.sscombo.setCurrentIndex(2) # Sagittal slice (R/L)
        self.ivl.set_focus([3, 0, 0, 0])
        self.processEvents()
        self.assertEquals(self.w.sscombo.currentText(), "Sagittal")
        self.assertAlmostEquals(float(model.item(0, 0).text()), np.mean(self.data_3d[3,:,:]), delta=0.01)
        self.assertAlmostEquals(float(model.item(1, 0).text()), np.median(self.data_3d[3,:,:]), delta=0.01)
        self.assertAlmostEquals(float(model.item(2, 0).text()), np.std(self.data_3d[3,:,:]), delta=0.01)
        self.assertAlmostEquals(float(model.item(3, 0).text()), np.min(self.data_3d[3,:,:]), delta=0.01)
        self.assertAlmostEquals(float(model.item(4, 0).text()), np.max(self.data_3d[3,:,:]), delta=0.01)
      
    def testSliceShowHide(self):
        self.harmless_click(self.w.butgenss)
        self.assertTrue(self.w.stats_table_ss.isVisible())
        self.harmless_click(self.w.butgenss)
        self.assertFalse(self.w.stats_table_ss.isVisible())
     
class MultiVoxelAnalysisTest(WidgetTest):

    def widget_class(self):
        return MultiVoxelAnalysis

    def testClick(self):
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        pt = (2, 2, 2, 0)
        self.ivl._pick(0, pt)
        self.processEvents()
        sig = self.data_4d[2,2,2,:]
        self.assertTrue(pt in self.w.plots)
        plot = self.w.plots[pt]
        for v1, v2 in zip(sig, plot.yvalues):
            self.assertAlmostEquals(v1, v2)
        self.assertTrue(plot in self.w.plot.items)

    def testMultiClick(self):
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        pt1 = (2, 2, 2, 0)
        sig1 = self.data_4d[2,2,2,:]
        self.ivl._pick(0, pt1)
        self.processEvents()
        pt2 = (3, 3, 3, 0)
        sig2 = self.data_4d[3,3,3,:]
        self.ivl._pick(0, pt2)
        self.processEvents()

        self.assertTrue(pt1 in self.w.plots)
        plot = self.w.plots[pt1]
        for v1, v2 in zip(sig1, plot.yvalues):
            self.assertAlmostEquals(v1, v2)
        self.assertTrue(plot in self.w.plot.items)

        self.assertTrue(pt2 in self.w.plots)
        plot = self.w.plots[pt2]
        for v1, v2 in zip(sig2, plot.yvalues):
            self.assertAlmostEquals(v1, v2)
        self.assertTrue(plot in self.w.plot.items)

    def testMultiClickChangeColor(self):
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        self.w.options.option("col").value = "red"
        self.processEvents()

        pt1 = (2, 2, 2, 0)
        self.ivl._pick(0, pt1)
        self.processEvents()

        self.assertTrue(pt1 in self.w.plots)
        plot = self.w.plots[pt1]
        self.assertEquals(plot.line_col, self.w.colors['red'])

        self.w.options.option("col").value = "blue"
        self.processEvents()

        pt2 = (3, 3, 3, 0)
        self.ivl._pick(0, pt2)
        self.processEvents()

        self.assertTrue(pt2 in self.w.plots)
        plot = self.w.plots[pt2]
        self.assertEquals(plot.line_col, self.w.colors['blue'])

    def testShowMeanCurves(self):
        """
        Select two points, show mean curves=On.
        """
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        self.w.options.option("col").value = "red"
        self.processEvents()
        self.w.options.option("mean").value = True
        self.processEvents()

        pt1 = (2, 2, 2, 0)
        self.ivl._pick(0, pt1)
        self.processEvents()

        self.assertTrue(pt1 in self.w.plots)
        plot = self.w.plots[pt1]

        pt2 = (3, 3, 3, 0)
        self.ivl._pick(0, pt2)
        self.processEvents()

        plot = self.w.mean_plots[self.w.colors["red"]]
        self.assertTrue(plot in self.w.plot.items)

        self.w.options.option("mean").value = False
        self.processEvents()
        self.assertEqual(len(plot.graphics_items), 0)

    def testShowIndividualCurves(self):
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        self.w.options.option("indiv").value = False
        self.processEvents()

        pt1 = (2, 2, 2, 0)
        self.ivl._pick(0, pt1)
        self.processEvents()

        pt2 = (3, 3, 3, 0)
        self.ivl._pick(0, pt2)
        self.processEvents()

        self.assertTrue(pt1 in self.w.plots)
        plot = self.w.plots[pt1]
        self.assertEqual(len(plot.graphics_items), 0)

        self.assertTrue(pt2 in self.w.plots)
        plot = self.w.plots[pt2]
        self.assertEqual(len(plot.graphics_items), 0)

        self.w.options.option("indiv").value = True
        self.processEvents()

        plot = self.w.plots[pt1]
        self.assertTrue(plot in self.w.plot.items)

        plot = self.w.plots[pt2]
        self.assertTrue(plot in self.w.plot.items)

class VoxelAnalysisTest(WidgetTest):

    def widget_class(self):
        return VoxelAnalysis

    def test3dDataOnly(self):
        self.ivm.add(self.data_4d, grid=self.grid, name="data_3d")
        pt = (2, 2, 2, 0)
        self.ivl._pick(0, pt)
        self.processEvents()
        self.assertFalse(self.error)

class MeasureWidgetTest(WidgetTest):

    def widget_class(self):
        return MeasureWidget

    def testDistance1d(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.harmless_click(self.w._dist_btn)
        self.processEvents()
        self.ivl._pick(0, (2, 2, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 2, 3, 0))
        self.processEvents()
        self.assertFalse(self.error)

        regex = re.compile("distance.*\s+(\d+(\.\d+)?)\s*mm", re.IGNORECASE)
        m = re.match(regex, self.w._label.text())
        self.assertTrue(m is not None)
        self.assertAlmostEquals(float(m.groups()[0]), 1.0, delta=0.01)

    def testDistance2d(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.harmless_click(self.w._dist_btn)
        self.processEvents()
        self.ivl._pick(0, (2, 2, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 3, 3, 0))
        self.processEvents()
        self.assertFalse(self.error)

        regex = re.compile("distance.*\s+(\d+(\.\d+)?)\s*mm", re.IGNORECASE)
        m = re.match(regex, self.w._label.text())
        self.assertTrue(m is not None)
        self.assertAlmostEquals(float(m.groups()[0]), 1.41421, delta=0.01)

    def testDistance3d(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.harmless_click(self.w._dist_btn)
        self.processEvents()
        self.ivl._pick(0, (2, 2, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (4, 3, 3, 0))
        self.processEvents()
        self.assertFalse(self.error)

        regex = re.compile("distance.*\s+(\d+(\.\d+)?)\s*mm", re.IGNORECASE)
        m = re.match(regex, self.w._label.text())
        self.assertTrue(m is not None)
        self.assertAlmostEquals(float(m.groups()[0]), 2.4495, delta=0.01)

    def testDistance3dNonSquiffy(self):
        """
        Distance with non-squiffy grid
        """
        affine = np.identity(4)
        affine[1, 1] = 2
        affine[2, 2] = 3
        grid = DataGrid(self.data_3d.shape, affine)

        self.ivm.add(self.data_3d, grid=grid, name="data_3d")
        self.harmless_click(self.w._dist_btn)
        self.processEvents()
        self.ivl._pick(0, (2, 2, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (4, 3, 3, 0))
        self.processEvents()
        self.assertFalse(self.error)

        regex = re.compile("distance.*\s+(\d+(\.\d+)?)\s*mm", re.IGNORECASE)
        m = re.match(regex, self.w._label.text())
        self.assertTrue(m is not None)
        self.assertAlmostEquals(float(m.groups()[0]), 4.1231, delta=0.01)

    def testAngleStraight(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.harmless_click(self.w._angle_btn)
        self.processEvents()
        self.ivl._pick(0, (2, 2, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 2, 3, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 2, 4, 0))
        self.processEvents()
        self.assertFalse(self.error)

        regex = re.compile("angle.*\s+(\d+(\.\d+)?).*", re.IGNORECASE)
        m = re.match(regex, self.w._label.text())
        self.assertTrue(m is not None)
        self.assertAlmostEquals(float(m.groups()[0]), 180, delta=0.01)

    def testAngleOrtho(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.harmless_click(self.w._angle_btn)
        self.processEvents()
        self.ivl._pick(0, (2, 2, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 2, 3, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 3, 3, 0))
        self.processEvents()
        self.assertFalse(self.error)

        regex = re.compile("angle.*\s+(\d+(\.\d+)?).*", re.IGNORECASE)
        m = re.match(regex, self.w._label.text())
        self.assertTrue(m is not None)
        self.assertAlmostEquals(float(m.groups()[0]), 90, delta=0.01)

    def testAngle45(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.harmless_click(self.w._angle_btn)
        self.processEvents()
        self.ivl._pick(0, (2, 2, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 3, 3, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 2, 3, 0))
        self.processEvents()
        self.assertFalse(self.error)

        regex = re.compile("angle.*\s+(\d+(\.\d+)?).*", re.IGNORECASE)
        m = re.match(regex, self.w._label.text())
        self.assertTrue(m is not None)
        self.assertAlmostEquals(float(m.groups()[0]), 45, delta=0.01)

    def testAngle45Squiffy(self):
        affine = np.identity(4)
        affine[1, 1] = 2
        affine[2, 2] = 3
        grid = DataGrid(self.data_3d.shape, affine)
        self.ivm.add(self.data_3d, grid=grid, name="data_3d")
        self.harmless_click(self.w._angle_btn)
        self.processEvents()
        self.ivl._pick(0, (2, 2, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (4, 3, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 3, 2, 0))
        self.processEvents()
        self.assertFalse(self.error)

        regex = re.compile("angle.*\s+(\d+(\.\d+)?).*", re.IGNORECASE)
        m = re.match(regex, self.w._label.text())
        self.assertTrue(m is not None)
        self.assertAlmostEquals(float(m.groups()[0]), 45, delta=0.01)

    def testAngleOrthoSquifffy(self):
        affine = np.identity(4)
        affine[1, 1] = 2
        affine[2, 2] = 3
        grid = DataGrid(self.data_3d.shape, affine)

        self.ivm.add(self.data_3d, grid=grid, name="data_3d")
        self.harmless_click(self.w._angle_btn)
        self.processEvents()
        self.ivl._pick(0, (2, 2, 2, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 2, 3, 0))
        self.processEvents()
        self.ivl._pick(0, (2, 3, 3, 0))
        self.processEvents()
        self.assertFalse(self.error)

        regex = re.compile("angle.*\s+(\d+(\.\d+)?).*", re.IGNORECASE)
        m = re.match(regex, self.w._label.text())
        self.assertTrue(m is not None)
        self.assertAlmostEquals(float(m.groups()[0]), 90, delta=0.01)

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Widget which displays list of all data loaded

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import print_function, division, absolute_import

import os

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets


from quantiphyse.data import save
from quantiphyse.gui.widgets import QpWidget, HelpButton, TextViewerDialog
from quantiphyse.utils import default_save_dir, get_icon, get_local_file
from quantiphyse import __contrib__, __acknowledge__

SUMMARY = "\nCreators: " + ", ".join([author for author in __contrib__]) \
          + "\nAcknowlegements: " + ", ".join([ack for ack in __acknowledge__])

class OverviewWidget(QpWidget):
    """
    QpWidget which displays welcome info and a list of current data sets
    """

    def __init__(self, **kwargs):
        super(OverviewWidget, self).__init__(name="Volumes", icon="volumes", desc="Overview of volumes loaded", group="DEFAULT", position=0, **kwargs)

    def init_ui(self):
        layout = QtGui.QVBoxLayout()

        hbox = QtGui.QHBoxLayout()
        pixmap = QtGui.QPixmap(get_icon("quantiphyse_75.png"))
        lpic = QtGui.QLabel(self)
        lpic.setPixmap(pixmap)
        hbox.addWidget(lpic)
        hbox.addStretch(1)
        help_btn = HelpButton(self)
        hbox.addWidget(help_btn)
        layout.addLayout(hbox)

        summary_label = QtGui.QLabel(SUMMARY)
        summary_label.setWordWrap(True)
        layout.addWidget(summary_label)

        box = QtGui.QGroupBox()
        hbox = QtGui.QHBoxLayout()
        box.setLayout(hbox)
        disc = QtGui.QLabel("<font size=2> Disclaimer: This software has been developed for research purposes only, and "
                            "should not be used as a diagnostic tool. The authors or distributors will not be "
                            "responsible for any direct, indirect, special, incidental, or consequential damages "
                            "arising of the use of this software. By using the this software you agree to this disclaimer."
                            "<p>"
                            "Please read the Quantiphyse License for more information")
        disc.setWordWrap(True)
        hbox.addWidget(disc, 10)
        license_btn = QtGui.QPushButton("License")
        license_btn.clicked.connect(self._view_license)
        hbox.addWidget(license_btn)
        layout.addWidget(box)

        self.data_list = DataListWidget(self)
        layout.addWidget(self.data_list)

        hbox = QtGui.QHBoxLayout()
        btn = QtGui.QPushButton("Rename")
        btn.clicked.connect(self._rename)
        hbox.addWidget(btn)
        btn = QtGui.QPushButton("Delete")
        btn.clicked.connect(self._delete)
        hbox.addWidget(btn)
        btn = QtGui.QPushButton("Save")
        btn.clicked.connect(self._save)
        hbox.addWidget(btn)
        btn = QtGui.QPushButton("Set as main data")
        btn.clicked.connect(self._set_main)
        hbox.addWidget(btn)
        btn = QtGui.QPushButton("Toggle ROI")
        btn.clicked.connect(self._toggle_roi)
        hbox.addWidget(btn)
        layout.addLayout(hbox)

        self.setLayout(layout)

    def _view_license(self):
        license_file = get_local_file("licence.md")
        with open(license_file, "r") as f:
            text = f.read()
        dlg = TextViewerDialog(self, "Quantiphyse License", text=text)
        dlg.exec_()

    def _delete(self):
        if self.data_list.selected is not None:
            name = self.data_list.selected.name
            ok = QtGui.QMessageBox.warning(self, "Delete data", "Delete '%s'?" % name,
                                           QtGui.QMessageBox.Ok | QtGui.QMessageBox.Cancel)
            if ok:
                self.ivm.delete(name)

    def _rename(self):
        if self.data_list.selected is not None:
            name = self.data_list.selected.name
            text, result = QtGui.QInputDialog.getText(self, "Renaming '%s'" % name, "New name",
                                                      QtGui.QLineEdit.Normal, name)
            if result:
                self.ivm.rename(name, text)

    def _save(self):
        if self.data_list.selected is not None:
            name = self.data_list.selected.name
            data = self.ivm.data[name]
            if hasattr(data, "fname") and data.fname is not None:
                fname = data.fname
            else:
                fname = os.path.join(default_save_dir(), name + ".nii")
            fname, _ = QtGui.QFileDialog.getSaveFileName(self, 'Save file', dir=fname,
                                                         filter="NIFTI files (*.nii *.nii.gz)")
            if fname != '':
                save(self.ivm.data[name], fname)
            else: # Cancelled
                pass

    def _set_main(self):
        if self.data_list.selected is not None:
            self.ivm.set_main_data(self.data_list.selected.name)

    def _toggle_roi(self):
        if self.data_list.selected is not None:
            self.data_list.selected.roi = not self.data_list.selected.roi
            self.ivm.sig_all_data.emit(list(self.ivm.data.keys()))

class DataListWidget(QtGui.QTableView):
    """
    Table showing loaded volumes
    """
    def __init__(self, parent):
        super(DataListWidget, self).__init__(parent)
        self.setStyleSheet("font-size: 10px; alternate-background-color: #6c6c6c;")

        self.ivm = parent.ivm
        self._selected = None
        self._roi_icon = QtGui.QIcon(get_icon("roi_data.png"))
        self._data_icon = QtGui.QIcon(get_icon("data.png"))
        self._main_icon = QtGui.QIcon(get_icon("main_data.png"))
        self._vis_icon = QtGui.QIcon(get_icon("visible.png"))
        self._main_vis_icon = QtGui.QIcon(get_icon("main_ovl.png"))

        self.model = QtGui.QStandardItemModel()
        self.setModel(self.model)
        self._update_list()

        self.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        self.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        self.setShowGrid(False)
        self.setTextElideMode(QtCore.Qt.ElideLeft)
        self.setAlternatingRowColors(True)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setDefaultSectionSize(self.verticalHeader().minimumSectionSize()+2)
        self.horizontalHeader().setVisible(True)
        self.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignLeft)

        self.clicked.connect(self._clicked)
        self.ivm.sig_all_data.connect(self._update_list)
        self.ivm.sig_main_data.connect(self._update_vis_icons)
        self.ivm.sig_current_data.connect(self._update_vis_icons)
        self.ivm.sig_current_roi.connect(self._update_vis_icons)

    @property
    def selected(self):
        """ Currently selected QpData """
        return self._selected

    def _get_table_items(self, data):
        fname = ""
        if hasattr(data, "fname") and data.fname is not None:
            fname = data.fname

        items = [
            QtGui.QStandardItem(""),
            QtGui.QStandardItem(data.name),
            QtGui.QStandardItem(fname)
        ]

        if fname:
            tooltip = fname
        else:
            tooltip = "Not saved to file"

        for item in items:
            item.setToolTip(tooltip)
        return items

    def _update_list(self):
        try:
            self.blockSignals(True)
            scroll_pos = self.verticalScrollBar().value()

            self.model.clear()
            self.model.setColumnCount(3)
            self.model.setHorizontalHeaderLabels(["", "Name", "File"])
            self.model.setHeaderData(0, QtCore.Qt.Horizontal, self._vis_icon, QtCore.Qt.DecorationRole)
            self.horizontalHeader().setResizeMode(0, QtGui.QHeaderView.ResizeToContents)
            self.horizontalHeader().setResizeMode(1, QtGui.QHeaderView.ResizeToContents)
            self.horizontalHeader().setResizeMode(2, QtGui.QHeaderView.Stretch)
            for row, name in enumerate(sorted(self.ivm.data.keys())):
                data = self.ivm.data.get(name)
                self.model.appendRow(self._get_table_items(data))

                index = self.model.index(row, 1)
                self.model.setData(index, self._roi_icon if data.roi else self._data_icon, QtCore.Qt.DecorationRole)
                self._update_vis_icon(row, data)

            self.verticalScrollBar().setValue(scroll_pos)
        finally:
            self.blockSignals(False)

    def _update_vis_icons(self):
        for row, name in enumerate(sorted(self.ivm.data.keys())):
            data = self.ivm.data.get(name)
            self._update_vis_icon(row, data)

    def _update_vis_icon(self, row, qpdata):
        is_main = qpdata == self.ivm.main
        is_cur = (qpdata == self.ivm.current_data or qpdata == self.ivm.current_roi)
        if is_main and is_cur:
            icon = self._main_vis_icon
        elif is_main:
            icon = self._main_icon
        elif is_cur:
            icon = self._vis_icon
        else:
            icon = None
        index = self.model.index(row, 0)
        self.model.setData(index, icon, QtCore.Qt.DecorationRole)

    def _selection(self, index):
        row, col = index.row(), index.column()
        name = self.model.item(row, 1).text()
        return row, col, name, self.ivm.data.get(name, None)

    def _clicked(self, index):
        row, col, name, data = self._selection(index)
        self._selected = data
        if col == 0:
            if data.roi:
                self.ivm.set_current_roi(data.name if data != self.ivm.current_roi else None)
            else:
                self.ivm.set_current_data(data.name if data != self.ivm.current_data else None)



"""
Quantiphyse - Package for data simulation

Copyright (c) 2013-2018 University of Oxford
"""
from .widgets import AddNoiseWidget, SimMotionWidget
from .processes import AddNoiseProcess, SimMotionProcess

QP_MANIFEST = {
    "widgets" : [AddNoiseWidget, SimMotionWidget],
    "processes" : [AddNoiseProcess, SimMotionProcess],
}



"""
Quantiphyse - Analysis processes for registration and motion correction

Registration is a little complex because it is a plugin which uses other
plugins - registration methods.

A registration method contains methods to do registration on 3d and 4d data,
and also motion correction. There are default base class implementations
of 4d registration and motion correction which simply use the 3d method, but
methods can override these defaults, e.g. the Flirt plugin can use McFlirt
to do motion correction

A registration method can also return the transform and apply it to other
data. Of course only transforms compatible with a registration method
can be used, e.g. rigid body vs fully deformable.

Copyright (c) 2013-2018 University of Oxford
"""

import sys
import logging
import traceback
import multiprocessing

import six
import numpy as np

from quantiphyse.data import NumpyData, QpData
from quantiphyse.data.extras import Extra
from quantiphyse.utils import get_plugins, set_local_file_path, QpException
from quantiphyse.processes import Process

LOG = logging.getLogger(__name__)

def get_reg_method(method_name):
    """
    Get a named registration method (case insensitive)
    """
    methods = get_plugins("reg-methods")
    LOG.debug("Known methods: %s", str(methods))
    for method_class in methods:
        method = method_class(None)
        if method.name.lower() == method_name.lower():
            return method
    return None

def _normalize_output(reg_data, output_data, output_suffix):
    if reg_data.roi:
        # This is not correct for multi-level ROIs - this would basically require support
        # from within the registration algorithm for roi (integer only) data
        data = np.rint(output_data.raw()).astype(np.int)
        output_data = NumpyData(data, grid=output_data.grid, name=output_data.name, roi=True)
    output_data.name = reg_data.name + output_suffix
    return output_data

def _reg_3d(worker_id, method, reg_data, ref_data, options, queue):
    """
    Register single volume data, may be more than one registration target
    """
    out_data = []
    log = "Running 3D registration\n\n"
    output_suffix = options.pop("output-suffix", "_reg")
    registered, transform, reg_log = method.reg_3d(reg_data[0], ref_data, dict(options), queue)
    registered = _normalize_output(reg_data[0], registered, output_suffix)
    log += reg_log
    out_data.append(registered)
    for idx, add_data in enumerate(reg_data[1:]):
        log += "\nApplying transformation to additional data %i\n\n" % (idx+1)
        registered, apply_log = method.apply_transform(add_data, transform, dict(options), queue)
        registered = _normalize_output(add_data, registered, output_suffix)
        log += apply_log
        out_data.append(registered)
    return worker_id, True, (out_data, transform, log)

def _reg_4d(worker_id, method, reg_data, ref_data, options, queue):
    """
    Register multi-volume data, can only be one registration target
    """
    if len(reg_data) > 1:
        raise QpException("Cannot have additional registration targets with 4D registration data")
    
    log = "Running 4D registration\n\n"
    output_suffix = options.pop("output-suffix", "_reg")
    out_data, transforms, reg_log = method.reg_4d(reg_data[0], ref_data, dict(options), queue)
    out_data = _normalize_output(reg_data[0], out_data, output_suffix)
    log += reg_log
    return worker_id, True, ([out_data], transforms, log)

def _reg_moco(worker_id, method, reg_data, ref_data, options, queue):
    """
    Motion correction mode
    """
    if len(reg_data) > 1:
        raise QpException("Cannot have additional registration targets with motion correction")
    
    log = "Running motion correction\n\n"
    output_suffix = options.pop("output-suffix", "_moco")
    out_data, transforms, moco_log = method.moco(reg_data[0], ref_data, dict(options), queue)
    out_data = _normalize_output(reg_data[0], out_data, output_suffix)
    log += moco_log
    return worker_id, True, ([out_data], transforms, log)

def _run_reg(worker_id, queue, method_name, mode, reg_data, ref_data, options):
    """
    Generic registration function for asynchronous process
    """
    try:
        set_local_file_path()
        method = get_reg_method(method_name)
        if method is None: 
            raise QpException("Unknown registration method: %s (known: %s)" % (method_name, str(get_plugins("reg-methods"))))

        if not reg_data:
            raise QpException("No registration data")
        elif mode == "moco":
            return _reg_moco(worker_id, method, reg_data, ref_data, options, queue)
        elif reg_data[0].ndim == 3:
            return _reg_3d(worker_id, method, reg_data, ref_data, options, queue)
        else:
            return _reg_4d(worker_id, method, reg_data, ref_data, options, queue)
    except:
        traceback.print_exc()
        return worker_id, False, sys.exc_info()[1]

class RegProcess(Process):
    """
    Asynchronous background process to run registration / motion correction
    """

    PROCESS_NAME = "Reg"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, worker_fn=_run_reg, **kwargs)

    def run(self, options):
        self.debug("Run")
        method_name = options.pop("method")
        mode = options.pop("mode", "reg")
        self._save_transform = options.pop("save-transform", None)

        # Registration data. Need to know the reg data grid and number of volumes so progress
        # and output data can be interpreted correctly
        regdata_name = options.pop("reg", options.pop("data", None))
        if not regdata_name:
            raise QpException("No registration data specified")
        elif regdata_name not in self.ivm.data:
            raise QpException("Registration data not found: %s" % regdata_name)
        regdata = self.ivm.data[regdata_name]
        reg_grid = regdata.grid
        reg_data = [regdata, ]
        self._output_name = options.pop("output-name", regdata.name + options.get("output-suffix", "_reg"))

        # Reference data. Defaults to same as reg data so MoCo can be
        # supported as self-registration
        refdata_name = options.pop("ref", regdata_name)
        ref_data = self.ivm.data[refdata_name]
        if ref_data.nvols > 1:
            refvol = options.pop("ref-vol", "median")
            if refvol == "median":
                refidx = int(ref_data.nvols/2)
            elif refvol == "mean":
                raise NotImplementedError("Not yet implemented")
            else:
                refidx = refvol
            ref_data = NumpyData(ref_data.volume(refidx), ref_data.grid, "refdata")

        # Additional registration targets can be specified which will be transformed
        # in the same way as the main registration data. 
        # 
        # Useful for masks defined on an unregistered volume.
        add_reg = options.pop("add-reg", options.pop("warp-rois", options.pop("warp-roi", ())))

        # Handle case where additional registation targets are given as a string
        if isinstance(add_reg, six.string_types):
            add_reg = (add_reg,)

        for name in add_reg:
            qpdata = self.ivm.data.get(name, None)
            if not qpdata:
                self.warn("Removing non-existant data set from additional registration list: %s" % name)
            else:
                reg_data.append(qpdata.resample(reg_grid, suffix=""))
        
        self.debug("Have %i registration targets" % len(reg_data))

        # Function input data must be passed as list of arguments for multiprocessingmethod = get_reg_method(method_name)
        options_pass = dict(options)
        options.clear()
        self.start_bg([method_name, mode, reg_data, ref_data, options_pass])

    def timeout(self, queue):
        if queue.empty(): return
        while not queue.empty():
            complete = queue.get()
        self.sig_progress.emit(complete)

    def finished(self, worker_output):
        """ Add output data to the IVM and set the log """
        if self.status == Process.SUCCEEDED:
            registered_data, transform, log = worker_output[0]
            # Output name applies to the registration input data
            registered_data[0].name = self._output_name
            self.log(log)

            first_roi, first_data = True, True
            for data in registered_data:
                if data.roi:
                    self.ivm.add(data, make_current=first_roi)
                    first_roi = False
                else:
                    self.ivm.add(data, make_current=first_data)
                    first_data = False

            if self._save_transform:
                if isinstance(transform, QpData):
                    self.ivm.add(transform, name=self._save_transform)
                elif isinstance(transform, Extra):
                    transform.name = self._save_transform
                    self.ivm.add_extra(self._save_transform, transform)
                else:
                    # FIXME what to do about moco - sequence of transforms?
                    pass

class ApplyTransformProcess(Process):
    """
    Asynchronous background process to run registration / motion correction
    """

    PROCESS_NAME = "ApplyTransform"

    def run(self, options):
        self.debug("Run")
        data = self.get_data(options)
        trans_name = options.pop("transform")
        output_name = options.pop("output-name", data.name + "_reg")

        transform = self.ivm.data.get(trans_name, None)
        if transform is None or "QpReg" not in transform.metadata:
            transform = self.ivm.extras.get(trans_name, None)

        if transform is None or "QpReg" not in transform.metadata:
            raise QpException("Transform not found: %s" % trans_name)

        method = get_reg_method(transform.metadata["QpReg"])
        if method is None:
            raise QpException("Registration method not found: %s" % transform.metadata["QpReg"])

        # Fake queue - we ignore progress reports as this process is not asynchronous
        queue = multiprocessing.Queue()

        self.log("Applying transformation to data: %s" % data.name)
        registered, apply_log = method.apply_transform(data, transform, dict(options), queue)
        registered = _normalize_output(data, registered, "_reg")
        self.log(apply_log)
        self.ivm.add(registered, name=output_name, make_current=True)
        
class MocoProcess(RegProcess):
    """
    Explicit motion correction process for backwards compat
    """

    PROCESS_NAME = "Moco"

    def run(self, options):
        options["mode"] = "moco"
        return RegProcess.run(self, options)



"""
Quantiphyse - Exception subclasses

Copyright (c) 2013-2018 University of Oxford
"""

class QpException(Exception):
    """ 
    Base class for quantiphyse-specific exceptions

    These are, by definition, exceptions which are expected under certain
    input conditions. They do not immediately imply a defect in the code
    (although a defect in the code might prompt one of these exceptions
    to be thrown). For example, invalid user input, malformed input files,
    etc, might all throw a QpException. No stack trace will be displayed
    for these exceptions, and when throwing the message/detail should
    be written so it can be understood by an end-user.

    By contrast, any exception within the normal Python exception hierarchy
    implies a bug in the code (even if the bug consists of 'need to throw a
    QpException instead). A stack trace will be displayed.
    
    Currently we do not define more specific subclasses although this
    might be useful in the future
    """

    def __init__(self, msg, detail=None):
        super(QpException, self).__init__(msg)
        self.msg = msg
        self.detail = detail

    def __str__(self):
        """ 
        String representation of exception
        
        Don't want QpException: in front of message as this type of exception
        is for the user to see
        """
        return self.msg



"""
Quantiphyse - Clustering process tests

Copyright (c) 2013-2018 University of Oxford
"""
import os
import unittest

from quantiphyse.processes import Process
from quantiphyse.test import ProcessTest

class KmeansProcessTest(ProcessTest):
    
    def testKmeans3d(self):
        yaml = """       
  - KMeans:
        data: testdata_3d
        n-clusters: 3
        output-name: test_clusters_3d
        
  - Save:
        test_clusters_3d:
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("test_clusters_3d" in self.ivm.data)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "test_clusters_3d.nii.gz")))

    def testKmeans4d(self):
        yaml = """
  - KMeans:
        data: testdata_4d
        n-clusters: 4
        n-pca: 5
        output-name: test_clusters_4d
        
  - Save:
        test_clusters_4d:
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("test_clusters_4d" in self.ivm.data)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "test_clusters_4d.nii.gz")))

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Basic data classes

Copyright (c) 2013-2018 University of Oxford
"""

from .qpdata import DataGrid, OrthoSlice, QpData, NumpyData
from .volume_management import ImageVolumeManagement
from .load_save import load, save
from .nifti import NiftiData

__all__ = ["DataGrid", "OrthoSlice", "QpData", "ImageVolumeManagement", 
           "NiftiData", "NumpyData", "load", "save"]



"""
Quantiphyse - Tests for the radial profile process

Copyright (c) 2013-2018 University of Oxford
"""
import os
import unittest

from quantiphyse.processes import Process
from quantiphyse.test import ProcessTest

class RadialProfileProcessTest(ProcessTest):
    
    def testRadialProfile(self):
        yaml = """ 
  - RadialProfile:
        data: data_3d
        centre: 2, 2, 2
        output-name: testdata_rp

  - SaveExtras: 
        testdata_rp: testdata_rp.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_rp" in self.ivm.extras)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "testdata_rp.tsv")))

if __name__ == '__main__':
    unittest.main()
